import time

import pytest
from cognite.client import CogniteClient
from cognite.client.data_classes import FileMetadataWrite, FunctionWrite

from cognite_toolkit._cdf_tk.client import ToolkitClient


@pytest.fixture()
def function_code_file_id(toolkit_client: ToolkitClient) -> int:
    """
    This fixture uploads a simple function code to CDF and returns the file ID.
    The function code is a simple Python function that prints the input data and function call info.
    """

    def handle(client: CogniteClient, data: dict, function_call_info: dict) -> dict:
        """
        [requirements]
        cognite-sdk>=7.78.0
        [/requirements]
        """
        print("Print statements will be shown in the logs.")
        print("Running with the following configuration:\n")
        return {
            "data": data,
            "functionInfo": function_call_info,
            "project": client.config.project,
        }

    name = "test_create_function_429"
    file = FileMetadataWrite(name=name, external_id="test_create_function_429")
    retrieved = toolkit_client.files.retrieve(external_id=file.external_id)
    if retrieved:
        return retrieved.id

    file_id = toolkit_client.functions._zip_and_upload_handle(handle, name=name, external_id=file.external_id)
    # Wait for the file to be available
    time.sleep(10)
    return file_id


class TestExtendedFunctions:
    def test_create_with_429_retry(self, toolkit_client: ToolkitClient, function_code_file_id: int) -> None:
        my_function = FunctionWrite(
            name="Test Function with 429 Retry",
            external_id="test_function_retry_429",
            file_id=function_code_file_id,
        )
        try:
            created = toolkit_client.functions.create_with_429_retry(my_function)
            assert created is not None
            assert created.external_id == my_function.external_id
            assert created.file_id == function_code_file_id
            assert created.name == my_function.name
        finally:
            toolkit_client.functions.delete_with_429_retry(
                external_id=[my_function.external_id], ignore_unknown_ids=True
            )



from contextlib import suppress
from time import sleep

import pandas as pd
import pytest
from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes import (
    ThreeDModelList,
    ThreeDModelWrite,
    ThreeDModelWriteList,
    TimeSeriesWrite,
    TimeSeriesWriteList,
)
from cognite.client.exceptions import CogniteAPIError

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.cruds import ContainerCRUD, TimeSeriesCRUD
from cognite_toolkit._cdf_tk.cruds._resource_cruds.three_d_model import ThreeDModelCRUD
from tests.test_integration.constants import RUN_UNIQUE_ID


@pytest.fixture(scope="session")
def integration_space(cognite_client: CogniteClient) -> dm.Space:
    space = dm.SpaceApply(
        space="integration_test_cdf_tk",
        name="CDF Toolkit Test Space",
        description="Space used for running integration test",
    )
    return cognite_client.data_modeling.spaces.apply(space)


class TestTimeSeriesLoader:
    def test_create_populate_count_drop_data(self, cognite_client: CogniteClient) -> None:
        timeseries = TimeSeriesWrite(
            external_id=f"test_create_populate_count_drop_data{RUN_UNIQUE_ID}", is_string=False
        )
        datapoints = pd.DataFrame(
            [{"timestamp": 0, timeseries.external_id: 0}, {"timestamp": 1, timeseries.external_id: 1}]
        ).set_index("timestamp")
        datapoints.index = pd.to_datetime(datapoints.index, unit="s")
        loader = TimeSeriesCRUD(cognite_client, None)
        ts_ids = [timeseries.external_id]

        try:
            created = loader.create(TimeSeriesWriteList([timeseries]))
            assert len(created) == 1

            assert loader.count(ts_ids) == 0
            cognite_client.time_series.data.insert_dataframe(datapoints)

            assert loader.count(ts_ids) == 2

            loader.drop_data(ts_ids)

            assert loader.count(ts_ids) == 0

            assert loader.delete(ts_ids) == 1

            assert not loader.retrieve(ts_ids)
        finally:
            cognite_client.time_series.delete(external_id=timeseries.external_id, ignore_unknown_ids=True)


@pytest.fixture(scope="function")
def node_container(cognite_client: CogniteClient, integration_space: dm.Space) -> dm.Container:
    container = dm.ContainerApply(
        space=integration_space.space,
        external_id="test_create_populate_count_drop_data",
        name="Test Container",
        description="Container used for running integration test",
        used_for="node",
        properties={"name": dm.ContainerProperty(dm.Text())},
    )
    return cognite_client.data_modeling.containers.apply(container)


@pytest.fixture(scope="function")
def edge_container(cognite_client: CogniteClient, integration_space: dm.Space) -> dm.Container:
    container = dm.ContainerApply(
        space=integration_space.space,
        external_id="test_create_populate_count_drop_data_edge",
        name="Test Container Edge",
        description="Container used for running integration test",
        used_for="edge",
        properties={"name": dm.ContainerProperty(dm.Text())},
    )
    return cognite_client.data_modeling.containers.apply(container)


class TestContainerLoader:
    # The DMS service is fairly unstable, so we need to rerun the tests if they fail.
    @pytest.mark.flaky(reruns=3, reruns_delay=10, only_rerun=["AssertionError", "CogniteAPIError"])
    def test_populate_count_drop_data_node_container(
        self, node_container: dm.Container, cognite_client: CogniteClient
    ) -> None:
        node = dm.NodeApply(
            space=node_container.space,
            external_id=f"test_create_populate_count_drop_data{RUN_UNIQUE_ID}",
            sources=[dm.NodeOrEdgeData(source=node_container.as_id(), properties={"name": "Anders"})],
        )
        container_id = [node_container.as_id()]

        loader = ContainerCRUD(cognite_client, None)

        try:
            assert loader.count(container_id) == 0

            cognite_client.data_modeling.instances.apply(nodes=[node])

            assert loader.count(container_id) == 1

            loader.drop_data(container_id)
            assert loader.count(container_id) == 0

            write_container = node_container.as_write()
            write_container.description = "Updated description"
            updated = loader.update(dm.ContainerApplyList([write_container]))
            assert len(updated) == 1
            if updated[0].description != write_container.description:
                # The API is not always consistent in returning the updated description,
                # so we need to retrieve the container to verify the update
                sleep(1)
                updated = loader.retrieve([node_container.as_id()])
            assert updated[0].description == write_container.description
        finally:
            loader.drop_data(container_id)

    # The DMS service is fairly unstable, so we need to rerun the tests if they fail.
    @pytest.mark.flaky(reruns=3, reruns_delay=10, only_rerun=["AssertionError", "CogniteAPIError"])
    def test_populate_count_drop_data_edge_container(
        self, edge_container: dm.Container, cognite_client: CogniteClient
    ) -> None:
        space = edge_container.space
        nodes = dm.NodeApplyList(
            [
                dm.NodeApply(
                    space=space,
                    external_id=f"test_create_populate_count_drop_data:start{RUN_UNIQUE_ID}",
                    sources=None,
                ),
                dm.NodeApply(
                    space=space,
                    external_id=f"test_create_populate_count_drop_data:end{RUN_UNIQUE_ID}",
                    sources=None,
                ),
            ]
        )
        edge = dm.EdgeApply(
            space=space,
            external_id=f"test_populate_count_drop_data_edge_container{RUN_UNIQUE_ID}",
            type=dm.DirectRelationReference(space, "test_edge_type"),
            start_node=(nodes[0].space, nodes[0].external_id),
            end_node=(nodes[1].space, nodes[1].external_id),
            sources=[dm.NodeOrEdgeData(source=edge_container.as_id(), properties={"name": "Anders"})],
        )
        container_id = [edge_container.as_id()]

        loader = ContainerCRUD(cognite_client, None)

        try:
            assert loader.count(container_id) == 0

            cognite_client.data_modeling.instances.apply(edges=[edge], nodes=nodes)

            assert loader.count(container_id) == 1

            loader.drop_data(container_id)
            assert loader.count(container_id) == 0

            write_container = edge_container.as_write()
            write_container.description = "Updated description"
            updated = loader.update(dm.ContainerApplyList([write_container]))
            assert len(updated) == 1
            if updated[0].description != write_container.description:
                # The API is not always consistent in returning the updated description,
                # so we need to retrieve the container to verify the update
                sleep(1)
                updated = loader.retrieve([write_container.as_id()])
            assert updated[0].description == write_container.description
        finally:
            cognite_client.data_modeling.instances.delete(nodes=nodes.as_ids(), edges=edge.as_id())


class Test3DModelLoader:
    def test_create_delete_model(self, toolkit_client: ToolkitClient) -> None:
        model = ThreeDModelWrite(
            name=f"tmp_test_create_delete_model_{RUN_UNIQUE_ID}",
            metadata={
                "description": "My description",
            },
        )

        loader = ThreeDModelCRUD(toolkit_client, None)

        missing = toolkit_client.iam.verify_capabilities(loader.get_required_capability(None, read_only=False))
        assert not missing, f"Missing capabilities: {missing}"

        created: ThreeDModelList | None = None
        try:
            created = loader.create(ThreeDModelWriteList([model]))
            assert len(created) == 1

            # Serialize and deserialize the model to get a copy
            update = ThreeDModelWrite.load(model.dump())
            update.metadata["new_key"] = "new_value"

            updated = loader.update(ThreeDModelWriteList([update]))
            assert len(updated) == 1
            assert updated[0].metadata["new_key"] == "new_value"

            delete_count = loader.delete([model.name])
            assert delete_count == 1
        finally:
            # Ensure that the model is deleted even if the test fails.
            if created:
                with suppress(CogniteAPIError):
                    toolkit_client.three_d.models.delete(created.as_ids())



from rich import print
from rich.panel import Panel

from ._base import ToolkitCommand


class InitCommand(ToolkitCommand):
    def execute(self) -> None:
        print(Panel("This command is deprecated. Use 'cdf modules init' instead."))



import json
from collections import Counter
from collections.abc import Iterator
from dataclasses import dataclass
from unittest.mock import patch

import httpx
import pytest
import respx

from cognite_toolkit._cdf_tk.client import ToolkitClientConfig
from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses
from cognite_toolkit._cdf_tk.utils.http_client import (
    ErrorDetails,
    FailedRequestItems,
    FailedRequestMessage,
    FailedResponse,
    FailedResponseItems,
    HTTPClient,
    HTTPMessage,
    ItemMessage,
    ItemsRequest,
    ParamRequest,
    SimpleBodyRequest,
    SuccessResponse,
    SuccessResponseItems,
)
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal
from tests.test_unit.utils import FakeCogniteResourceGenerator


@pytest.fixture
def rsps() -> Iterator[respx.MockRouter]:
    with respx.mock() as rsps:
        yield rsps


@pytest.fixture
def http_client(toolkit_config: ToolkitClientConfig) -> Iterator[HTTPClient]:
    with HTTPClient(toolkit_config) as client:
        yield client


@pytest.fixture
def http_client_one_retry(toolkit_config: ToolkitClientConfig) -> Iterator[HTTPClient]:
    with HTTPClient(toolkit_config, max_retries=1) as client:
        yield client


@pytest.mark.usefixtures("disable_pypi_check")
class TestHTTPClient:
    def test_get_request(self, rsps: respx.MockRouter, http_client: HTTPClient) -> None:
        rsps.get("https://example.com/api/resource").respond(json={"key": "value"}, status_code=200)
        results = http_client.request(
            ParamRequest(endpoint_url="https://example.com/api/resource", method="GET", parameters={"query": "test"})
        )
        assert len(results) == 1
        response = results[0]
        assert isinstance(response, SuccessResponse)
        assert response.status_code == 200
        assert response.body == '{"key":"value"}'
        assert rsps.calls[-1].request.url == "https://example.com/api/resource?query=test"

    @pytest.mark.usefixtures("disable_gzip")
    def test_post_request(self, rsps: respx.MockRouter, http_client: HTTPClient) -> None:
        rsps.post("https://example.com/api/resource").respond(json={"id": 123, "status": "created"}, status_code=201)
        results = http_client.request(
            SimpleBodyRequest(
                endpoint_url="https://example.com/api/resource", method="POST", body_content={"name": "new resource"}
            )
        )
        assert len(results) == 1
        response = results[0]
        assert isinstance(response, SuccessResponse)
        assert response.status_code == 201
        assert response.body == '{"id":123,"status":"created"}'
        assert rsps.calls[-1].request.content == b'{"name":"new resource"}'

    @pytest.mark.usefixtures("disable_gzip")
    @pytest.mark.parametrize(
        "bad_body,error",
        [
            pytest.param({"name": bytes(10)}, "can't be serialized by the JSON encoder", id="bytes value"),
            pytest.param(
                {"values": [float("nan")], "other": float("inf")},
                "Out of range float values are not JSON compliant",
                id="nan and inf",
            ),
        ],
    )
    def test_post_request_bad_json(self, bad_body: dict[str, JsonVal], error: str, http_client: HTTPClient) -> None:
        results = http_client.request(
            SimpleBodyRequest(
                endpoint_url="https://example.com/api/resource",
                method="POST",
                body_content=bad_body,
            )
        )
        assert len(results) == 1
        response = results[0]
        assert isinstance(response, FailedRequestMessage)
        assert error in response.error

    def test_failed_request(self, rsps: respx.MockRouter, http_client: HTTPClient) -> None:
        rsps.get("https://example.com/api/resource").respond(json={"error": "bad request"}, status_code=400)
        results = http_client.request(
            ParamRequest(endpoint_url="https://example.com/api/resource", method="GET", parameters={"query": "fail"})
        )
        assert len(results) == 1
        response = results[0]
        assert isinstance(response, FailedResponse)
        assert response.status_code == 400
        assert response.error.message == "bad request"

    @pytest.mark.usefixtures("disable_gzip")
    def test_retry_then_success(self, rsps: respx.MockRouter, http_client: HTTPClient) -> None:
        url = "https://example.com/api/resource"
        rsps.get(url).respond(json={"error": "service unavailable"}, status_code=503)
        rsps.get(url).respond(json={"key": "value"}, status_code=200)
        results = http_client.request_with_retries(ParamRequest(endpoint_url=url, method="GET"))
        assert len(results) == 1
        response = results[0]
        assert isinstance(response, SuccessResponse)
        assert response.status_code == 200
        assert response.body == '{"key":"value"}'

    def test_retry_exhausted(self, http_client_one_retry: HTTPClient, rsps: respx.MockRouter) -> None:
        client = http_client_one_retry
        for _ in range(2):
            rsps.get("https://example.com/api/resource").respond(
                json={"error": {"message": "service unavailable", "code": 503}}, status_code=503
            )
        with patch("time.sleep"):  # Patch sleep to speed up the test
            results = client.request_with_retries(
                ParamRequest(endpoint_url="https://example.com/api/resource", method="GET")
            )

        assert len(results) == 1
        response = results[0]
        assert isinstance(response, FailedResponse)
        assert response.status_code == 503
        assert response.error.message == "service unavailable"

    def test_connection_error(self, http_client_one_retry: HTTPClient, rsps: respx.MockRouter) -> None:
        http_client = http_client_one_retry
        rsps.get("http://nonexistent.domain/api/resource").mock(
            side_effect=httpx.ConnectError("Simulated connection error")
        )
        with patch(f"{HTTPClient.__module__}.time"):
            # Patch time to avoid actual sleep
            results = http_client.request_with_retries(
                ParamRequest(endpoint_url="http://nonexistent.domain/api/resource", method="GET")
            )
        response = results[0]
        assert len(results) == 1
        assert isinstance(response, FailedRequestMessage)
        assert "RequestException after 1 attempts (connect error): Simulated connection error" == response.error

    def test_read_timeout_error(self, http_client_one_retry: HTTPClient, rsps: respx.MockRouter) -> None:
        http_client = http_client_one_retry
        rsps.get("https://example.com/api/resource").mock(side_effect=httpx.ReadTimeout("Simulated read timeout"))
        bad_request = ParamRequest(endpoint_url="https://example.com/api/resource", method="GET")
        with patch(f"{HTTPClient.__module__}.time"):
            # Patch time to avoid actual sleep
            results = http_client.request_with_retries(bad_request)
        response = results[0]
        assert len(results) == 1
        assert isinstance(response, FailedRequestMessage)
        assert "RequestException after 1 attempts (read error): Simulated read timeout" == response.error

    def test_zero_retries(self, toolkit_config: ToolkitClientConfig, rsps: respx.MockRouter) -> None:
        client = HTTPClient(toolkit_config, max_retries=0)
        rsps.get("https://example.com/api/resource").respond(json={"error": "service unavailable"}, status_code=503)
        results = client.request_with_retries(
            ParamRequest(endpoint_url="https://example.com/api/resource", method="GET")
        )
        assert len(results) == 1
        response = results[0]
        assert isinstance(response, FailedResponse)
        assert response.status_code == 503
        assert response.error.message == "service unavailable"
        assert len(rsps.calls) == 1

    def test_raise_if_already_retied(self, http_client_one_retry: HTTPClient) -> None:
        http_client = http_client_one_retry
        bad_request = ParamRequest(endpoint_url="https://example.com/api/resource", method="GET", status_attempt=3)
        with pytest.raises(RuntimeError, match=r"RequestMessage has already been attempted 3 times."):
            http_client.request_with_retries(bad_request)

    def test_error_text(self, http_client: HTTPClient, rsps: respx.MockRouter) -> None:
        rsps.get("https://example.com/api/resource").respond(json={"message": "plain_text"}, status_code=401)
        results = http_client.request(ParamRequest(endpoint_url="https://example.com/api/resource", method="GET"))
        assert len(results) == 1
        response = results[0]
        assert isinstance(response, FailedResponse)
        assert response.status_code == 401
        assert response.error.message == '{"message":"plain_text"}'

    def test_request_alpha(self, http_client: HTTPClient, rsps: respx.MockRouter) -> None:
        rsps.get("https://example.com/api/alpha/endpoint").respond(json={"key": "value"}, status_code=200)
        results = http_client.request(
            ParamRequest(
                endpoint_url="https://example.com/api/alpha/endpoint",
                method="GET",
                parameters={"query": "test"},
                api_version="alpha",
            )
        )
        assert len(results) == 1
        response = results[0]
        assert isinstance(response, SuccessResponse)
        assert response.status_code == 200
        assert rsps.calls[-1].request.headers["cdf-version"] == "alpha"


@dataclass
class MyRequestItemID:
    item: dict[str, JsonVal]

    def dump(self) -> JsonVal:
        return self.item

    def as_id(self) -> int:
        return self.item["id"]


@dataclass
class MyRequestItemExternalID:
    item: dict[str, JsonVal]

    def dump(self) -> JsonVal:
        return self.item

    def as_id(self) -> str:
        return self.item["externalId"]


@pytest.mark.usefixtures("disable_pypi_check")
class TestHTTPClientItemRequests:
    @pytest.mark.usefixtures("disable_gzip")
    def test_request_with_items_happy_path(self, http_client: HTTPClient, rsps: respx.MockRouter) -> None:
        rsps.post("https://example.com/api/resource").respond(
            json={"items": [{"id": 1, "value": 42}, {"id": 2, "value": 43}]},
            status_code=200,
        )
        items = [MyRequestItemID({"name": "item1", "id": 1}), MyRequestItemID({"name": "item2", "id": 2})]
        results = http_client.request(
            ItemsRequest[int](
                endpoint_url="https://example.com/api/resource",
                method="POST",
                items=items,
                extra_body_fields={"autoCreateDirectRelations": True},
            )
        )
        assert results == [
            SuccessResponseItems(
                status_code=200, ids=[1, 2], body='{"items":[{"id":1,"value":42},{"id":2,"value":43}]}'
            )
        ]
        assert len(rsps.calls) == 1
        assert json.loads(rsps.calls[0].request.content) == {
            "items": [{"name": "item1", "id": 1}, {"name": "item2", "id": 2}],
            "autoCreateDirectRelations": True,
        }

    @pytest.mark.usefixtures("disable_gzip")
    def test_request_with_items_issues(self, http_client: HTTPClient, rsps: respx.MockRouter) -> None:
        response_items = [
            {"externalId": "success", "data": 123},
        ]

        def server_callback(request: httpx.Request) -> httpx.Response:
            # Check request body content
            body_content = request.content.decode() if request.content else ""
            if "fail" in body_content:
                return httpx.Response(400, json={"error": "Item failed"})
            elif "success" in body_content:
                return httpx.Response(200, json={"items": response_items})
            else:
                return httpx.Response(200, json={"items": []})

        rsps.post("https://example.com/api/resource").mock(side_effect=server_callback)

        request_items = [
            MyRequestItemExternalID({"externalId": "success"}),
            MyRequestItemExternalID({"externalId": "fail"}),
        ]
        results = http_client.request_with_retries(
            ItemsRequest[str](
                endpoint_url="https://example.com/api/resource",
                method="POST",
                items=request_items,
            )
        )
        assert results == [
            SuccessResponseItems(
                status_code=200, ids=["success"], body='{"items":[{"externalId":"success","data":123}]}'
            ),
            FailedResponseItems(
                status_code=400,
                ids=["fail"],
                error=ErrorDetails(message="Item failed", code=400),
                body='{"error":"Item failed"}',
            ),
        ]
        assert len(rsps.calls) == 3  # Three requests made
        first, second, third = rsps.calls
        # First call will fail, and split into 1 item + 2 items
        assert json.loads(first.request.content)["items"] == [
            {"externalId": "success"},
            {"externalId": "fail"},
        ]
        # Second succeeds with 1 item.
        assert json.loads(second.request.content)["items"] == [{"externalId": "success"}]
        # Third fails with 1 item.
        assert json.loads(third.request.content)["items"] == [{"externalId": "fail"}]

    def test_request_all_item_fail(self, http_client: HTTPClient, rsps: respx.MockRouter) -> None:
        rsps.post("https://example.com/api/resource").respond(
            json={"error": "Unauthorized"},
            status_code=401,
        )
        items = [MyRequestItemID({"name": "item1", "id": 1}), MyRequestItemID({"name": "item2", "id": 2})]
        results = http_client.request(
            ItemsRequest[int](
                endpoint_url="https://example.com/api/resource",
                method="POST",
                items=items,
            )
        )
        assert results == [
            FailedResponseItems(
                status_code=401,
                ids=[1, 2],
                error=ErrorDetails(message="Unauthorized", code=401),
                body='{"error":"Unauthorized"}',
            ),
        ]
        assert len(rsps.calls) == 1  # Only one request made

    def test_request_no_items_response(self, http_client: HTTPClient, rsps: respx.MockRouter) -> None:
        rsps.post("https://example.com/api/resource/delete").respond(status_code=200)
        items = [MyRequestItemID({"id": 1}), MyRequestItemID({"id": 2})]
        results = http_client.request(
            ItemsRequest[int](
                endpoint_url="https://example.com/api/resource/delete",
                method="POST",
                items=items,
            )
        )
        assert results == [
            SuccessResponseItems(status_code=200, ids=[1, 2], body=""),
        ]

    def test_timeout_error(self, http_client_one_retry: HTTPClient, rsps: respx.MockRouter) -> None:
        client = http_client_one_retry
        rsps.post("https://example.com/api/resource").mock(side_effect=httpx.ReadTimeout("Simulated timeout error"))
        with patch("time.sleep"):
            results = client.request_with_retries(
                ItemsRequest[int](
                    endpoint_url="https://example.com/api/resource",
                    method="POST",
                    items=[MyRequestItemID({"id": 1})],
                )
            )
        assert results == [
            FailedRequestItems(ids=[1], error="RequestException after 1 attempts (read error): Simulated timeout error")
        ]

    @pytest.mark.usefixtures("disable_gzip")
    def test_early_failure(self, http_client_one_retry: HTTPClient, rsps: respx.MockRouter) -> None:
        client = http_client_one_retry
        rsps.post("https://example.com/api/resource").respond(
            json={"error": "Server error"},
            status_code=400,
        )
        with patch("time.sleep"):
            results = client.request_with_retries(
                ItemsRequest[int](
                    endpoint_url="https://example.com/api/resource",
                    method="POST",
                    items=[MyRequestItemID({"id": i}) for i in range(1000)],
                    max_failures_before_abort=5,
                )
            )
        assert len(rsps.calls) == 5
        actual_items_per_request = [len(json.loads(call.request.content)["items"]) for call in rsps.calls]
        assert actual_items_per_request == [1000, 500, 500, 250, 250]  # Splits in half each time
        failures = Counter([type(result) for result in results if isinstance(result, ItemMessage) for _ in result.ids])
        assert failures == {
            FailedResponseItems: 250,  # 250 items keeps the original error message.
            FailedRequestItems: 750,  # 750 items get the early abort message.
        }

    @pytest.mark.usefixtures("disable_gzip")
    def test_abort_on_first_failure(self, http_client: HTTPClient, rsps: respx.MockRouter) -> None:
        client = http_client
        rsps.post("https://example.com/api/resource").respond(
            json={"error": "Server error"},
            status_code=400,
        )
        results = client.request_with_retries(
            ItemsRequest[int](
                endpoint_url="https://example.com/api/resource",
                method="POST",
                items=[MyRequestItemID({"id": i}) for i in range(1000)],
                max_failures_before_abort=1,
            )
        )
        actual_failure_types = Counter([type(results) for results in results])
        assert actual_failure_types == {FailedResponseItems: 1}
        assert len(rsps.calls) == 1
        actual_items_per_request = [len(json.loads(call.request.content)["items"]) for call in rsps.calls]
        assert actual_items_per_request == [1000]

    @pytest.mark.usefixtures("disable_gzip")
    def test_abort_on_second_failure(self, http_client: HTTPClient, rsps: respx.MockRouter) -> None:
        client = http_client
        rsps.post("https://example.com/api/resource").respond(
            json={"error": "Server error"},
            status_code=400,
        )
        results = client.request_with_retries(
            ItemsRequest[int](
                endpoint_url="https://example.com/api/resource",
                method="POST",
                items=[MyRequestItemID({"id": i}) for i in range(1000)],
                max_failures_before_abort=2,
            )
        )
        actual_failure_types = Counter(
            [type(results) for results in results if isinstance(results, ItemMessage) for _ in results.ids]
        )
        assert actual_failure_types == {FailedResponseItems: 500, FailedRequestItems: 500}
        assert len(rsps.calls) == 2
        actual_items_per_request = [len(json.loads(call.request.content)["items"]) for call in rsps.calls]
        assert actual_items_per_request == [1000, 500]

    @pytest.mark.usefixtures("disable_gzip")
    def test_never_abort_on_failure(self, http_client: HTTPClient, rsps: respx.MockRouter) -> None:
        client = http_client
        rsps.post("https://example.com/api/resource").respond(
            json={"error": "Server error"},
            status_code=400,
        )
        results = client.request_with_retries(
            ItemsRequest[int](
                endpoint_url="https://example.com/api/resource",
                method="POST",
                items=[MyRequestItemID({"id": i}) for i in range(100)],
                max_failures_before_abort=-1,  # Never abort
            )
        )
        actual_failure_types = Counter(
            [type(results) for results in results if isinstance(results, ItemMessage) for _ in results.ids]
        )
        assert actual_failure_types == {FailedResponseItems: 100}
        assert len(rsps.calls) == 199

    @pytest.mark.usefixtures("disable_gzip")
    def test_failing_3_items(self, http_client_one_retry: HTTPClient, rsps: respx.MockRouter) -> None:
        client = http_client_one_retry

        def dislike_942_112_and_547(request: httpx.Request) -> httpx.Response:
            # Check request body content
            body_content = request.content.decode() if request.content else ""
            for no in ["942", "112", "547"]:
                if no in body_content:
                    return httpx.Response(400, json={"error": f"Item {no} is not allowed"})

            # Parse the request body to create response items
            try:
                body_data = json.loads(body_content)
                items = body_data.get("items", [])
                response_items = [{"id": item["id"], "status": "ok"} for item in items]
                return httpx.Response(200, json={"items": response_items})
            except (json.JSONDecodeError, KeyError):
                return httpx.Response(200, json={"items": []})

        rsps.post("https://example.com/api/resource").mock(side_effect=dislike_942_112_and_547)
        with patch("time.sleep"):
            results = client.request_with_retries(
                ItemsRequest[int](
                    endpoint_url="https://example.com/api/resource",
                    method="POST",
                    items=[MyRequestItemID({"id": i}) for i in range(1000)],
                    max_failures_before_abort=30,
                )
            )
        failures = Counter(
            [type(results) for results in results if isinstance(results, ItemMessage) for _ in results.ids]
        )
        assert failures == {FailedResponseItems: 3, SuccessResponseItems: 997}


class TestHTTPMessage:
    @pytest.mark.parametrize("message_cls", get_concrete_subclasses(HTTPMessage))
    def test_dump_http_message(self, message_cls: type[HTTPMessage]) -> None:
        message = FakeCogniteResourceGenerator(seed=42).create_instance(message_cls)

        dumped = message.dump()
        assert isinstance(dumped, dict)
        assert dumped["type"] == message_cls.__name__
        try:
            json.dumps(dumped)
        except (TypeError, ValueError) as e:
            pytest.fail(f"Dumped data is not valid JSON: {e}")



from typing import TYPE_CHECKING

from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitError,
)

if TYPE_CHECKING:
    from sentry_sdk.types import Event as SentryEvent
    from sentry_sdk.types import Hint as SentryHint


def sentry_exception_filter(event: "SentryEvent", hint: "SentryHint") -> "SentryEvent | None":
    if "exc_info" in hint:
        _exc_type, exc_value, _tb = hint["exc_info"]
        # Returning None prevents the event from being sent to Sentry
        if isinstance(exc_value, ToolkitError):
            return None
    return event



from collections.abc import Iterable, Sequence
from typing import Any

from cognite.client.data_classes import Annotation, AnnotationFilter

from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._asset_centric import FileMetadataIO
from ._base import Page, StorageIO
from .selectors import AssetCentricSelector


class FileAnnotationIO(StorageIO[AssetCentricSelector, Annotation]):
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    CHUNK_SIZE = 1000
    BASE_SELECTOR = AssetCentricSelector

    MISSING_ID = "<MISSING_RESOURCE_ID>"

    def as_id(self, item: Annotation) -> str:
        project = item._cognite_client.config.project
        return f"INTERNAL_ID_project_{project}_{item.id!s}"

    def stream_data(self, selector: AssetCentricSelector, limit: int | None = None) -> Iterable[Page]:
        total = 0
        for file_chunk in FileMetadataIO(self.client).stream_data(selector, None):
            # Todo Support pagination. This is missing in the SDK.
            results = self.client.annotations.list(
                filter=AnnotationFilter(
                    annotated_resource_type="file",
                    annotated_resource_ids=[{"id": file_metadata.id} for file_metadata in file_chunk.items],
                )
            )
            if limit is not None and total + len(results) > limit:
                results = results[: limit - total]

            for chunk in chunker_sequence(results, self.CHUNK_SIZE):
                yield Page(worker_id="main", items=chunk)
                total += len(chunk)
            if limit is not None and total >= limit:
                break

    def count(self, selector: AssetCentricSelector) -> int | None:
        """There is no efficient way to count annotations in CDF."""
        return None

    def data_to_json_chunk(
        self, data_chunk: Sequence[Annotation], selector: AssetCentricSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        files_ids: set[int] = set()
        for item in data_chunk:
            if item.annotated_resource_type == "file" and item.annotated_resource_id is not None:
                files_ids.add(item.annotated_resource_id)
            if file_id := self._get_file_id(item.data):
                files_ids.add(file_id)
        self.client.lookup.files.external_id(list(files_ids))  # Preload file external IDs
        asset_ids = {asset_id for item in data_chunk if (asset_id := self._get_asset_id(item.data))}
        self.client.lookup.assets.external_id(list(asset_ids))  # Preload asset external IDs
        return [self.dump_annotation_to_json(item) for item in data_chunk]

    def dump_annotation_to_json(self, annotation: Annotation) -> dict[str, JsonVal]:
        """Dump annotations to a list of JSON serializable dictionaries.

        Args:
            annotation: The annotations to dump.

        Returns:
            A list of JSON serializable dictionaries representing the annotations.
        """
        dumped = annotation.as_write().dump()
        if isinstance(annotated_resource_id := dumped.pop("annotatedResourceId", None), int):
            external_id = self.client.lookup.files.external_id(annotated_resource_id)
            dumped["annotatedResourceExternalId"] = self.MISSING_ID if external_id is None else external_id

        if isinstance(data := dumped.get("data"), dict):
            if isinstance(file_ref := data.get("fileRef"), dict) and isinstance(file_ref.get("id"), int):
                external_id = self.client.lookup.files.external_id(file_ref.pop("id"))
                file_ref["externalId"] = self.MISSING_ID if external_id is None else external_id
            if isinstance(asset_ref := data.get("assetRef"), dict) and isinstance(asset_ref.get("id"), int):
                external_id = self.client.lookup.assets.external_id(asset_ref.pop("id"))
                asset_ref["externalId"] = self.MISSING_ID if external_id is None else external_id
        return dumped

    @classmethod
    def _get_file_id(cls, data: dict[str, Any]) -> int | None:
        file_ref = data.get("fileRef")
        if isinstance(file_ref, dict):
            id_ = file_ref.get("id")
            if isinstance(id_, int):
                return id_
        return None

    @classmethod
    def _get_asset_id(cls, data: dict[str, Any]) -> int | None:
        asset_ref = data.get("assetRef")
        if isinstance(asset_ref, dict):
            id_ = asset_ref.get("id")
            if isinstance(id_, int):
                return id_
        return None



from typing import Literal

from pydantic import Field

from .base import BaseModelResource, ToolkitResource


class LabelRefYAML(BaseModelResource):
    externalId: str = Field(
        description="The external ID of the label.",
        max_length=255,
    )


class RelationshipYAML(ToolkitResource):
    external_id: str = Field(
        description="The external ID provided by the client.",
        max_length=255,
    )
    source_external_id: str = Field(
        description="The external ID of the resource that is the relationship source.",
    )
    source_type: Literal["asset", "timeSeries", "file", "event", "sequence"] = Field(
        description="The resource type of the relationship source. Must be one of the specified values."
    )
    target_external_id: str = Field(
        description="The external ID of the resource that is the relationship target.",
    )
    target_type: Literal["asset", "timeSeries", "file", "event", "sequence"] = Field(
        description="The resource type of the relationship target. Must be one of the specified values."
    )
    start_time: int | None = Field(
        None,
        description="The time, in milliseconds since Jan. 1, 1970, when the relationship became active. If there is no startTime, the relationship is active from the beginning of time until endTime.",
        ge=0,
    )
    end_time: int | None = Field(
        None,
        description="The time, in milliseconds since Jan. 1, 1970, when the relationship became inactive. If there is no endTime, the relationship is active until the end of time.",
        ge=0,
    )
    confidence: float | None = Field(
        None, description="A number between 0 and 1 that indicates the confidence in the relationship.", ge=0, le=1
    )
    data_set_external_id: str | None = Field(
        None, description="The external ID of the data set that the relationship belongs to.", max_length=255
    )
    labels: list[LabelRefYAML] | None = Field(
        None, description="A list of labels that the relationship belongs to.", max_length=10
    )



from typing import Any

import typer
from rich import print

from ._download_app import DownloadApp
from ._purge import PurgeApp
from ._upload_app import UploadApp


class DataApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.add_typer(DownloadApp(*args, **kwargs), name="download")
        self.add_typer(UploadApp(*args, **kwargs), name="upload")
        self.add_typer(PurgeApp(*args, **kwargs), name="purge")

    @staticmethod
    def main(ctx: typer.Context) -> None:
        """Commands to work with data download/upload/purge."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf data --help[/] for more information.")
        return None



from collections.abc import Iterable

import pytest

from cognite_toolkit._cdf_tk.resource_classes.views import ViewYAML
from tests.test_unit.utils import find_resources


def view_property_type_cases() -> Iterable:
    view_property_types = [
        {
            "CreateViewProperty": {
                "space": "my_space",
                "externalId": "my_view",
                "version": "1",
                "name": "My View",
                "description": "A test view with create view property",
                "properties": {
                    "textField": {
                        "name": "Text Field",
                        "description": "A text field from container",
                        "container": {"type": "container", "space": "my_space", "externalId": "my_container"},
                        "containerPropertyIdentifier": "textField",
                    }
                },
            }
        },
        {
            "SingleEdgeConnection": {
                "space": "my_space",
                "externalId": "edge_view",
                "version": "1",
                "properties": {
                    "connectedTo": {
                        "connectionType": "single_edge_connection",
                        "source": {"type": "view", "space": "my_space", "externalId": "source_view", "version": "1"},
                        "type": {"space": "my_space", "externalId": "edge_type"},
                        "direction": "outwards",
                    }
                },
            }
        },
        {
            "MultiEdgeConnection": {
                "space": "my_space",
                "externalId": "multi_edge_view",
                "version": "1",
                "properties": {
                    "manyConnections": {
                        "connectionType": "multi_edge_connection",
                        "source": {"type": "view", "space": "my_space", "externalId": "source_view", "version": "1"},
                        "type": {"space": "my_space", "externalId": "edge_type"},
                        "direction": "inwards",
                    }
                },
            }
        },
        {
            "SingleReverseDirectRelation": {
                "space": "my_space",
                "externalId": "reverse_view",
                "version": "1",
                "properties": {
                    "reverseConnection": {
                        "connectionType": "single_reverse_direct_relation",
                        "source": {"type": "view", "space": "my_space", "externalId": "source_view", "version": "1"},
                        "through": {
                            "source": {"type": "container", "space": "my_space", "externalId": "relation_container"},
                            "identifier": "connectedTo",
                        },
                    }
                },
            }
        },
        {
            "MultiReverseDirectRelation": {
                "space": "my_space",
                "externalId": "multi_reverse_view",
                "version": "1",
                "properties": {
                    "multiReverseConnection": {
                        "connectionType": "multi_reverse_direct_relation",
                        "source": {"type": "view", "space": "my_space", "externalId": "source_view", "version": "1"},
                        "through": {
                            "source": {
                                "type": "view",
                                "space": "my_space",
                                "externalId": "relation_view",
                                "version": "1",
                            },
                            "identifier": "connectedTo",
                        },
                    }
                },
            }
        },
    ]
    yield from (pytest.param(next(iter(pt.values())), id=next(iter(pt.keys()))) for pt in view_property_types)


def view_with_implements_cases() -> Iterable:
    view_implements = [
        {
            "ImplementsView": {
                "space": "my_space",
                "externalId": "implements_view",
                "version": "1",
                "implements": [{"type": "view", "space": "my_space", "externalId": "base_view", "version": "1"}],
                "properties": {
                    "name": {
                        "container": {"type": "container", "space": "my_space", "externalId": "my_container"},
                        "containerPropertyIdentifier": "name",
                    }
                },
            }
        },
        {
            "ImplementsMultipleViews": {
                "space": "my_space",
                "externalId": "multi_implements_view",
                "version": "1",
                "implements": [
                    {"type": "view", "space": "my_space", "externalId": "base_view1", "version": "1"},
                    {"type": "view", "space": "my_space", "externalId": "base_view2", "version": "1"},
                ],
                "properties": {
                    "identifier": {
                        "container": {"type": "container", "space": "my_space", "externalId": "my_container"},
                        "containerPropertyIdentifier": "identifier",
                    }
                },
            }
        },
    ]
    yield from (pytest.param(next(iter(i.values())), id=next(iter(i.keys()))) for i in view_implements)


def view_with_filters_cases() -> Iterable:
    view_filters = [
        {
            "SimpleFilter": {
                "space": "my_space",
                "externalId": "filtered_view",
                "version": "1",
                "filter": {"equals": {"property": ["status"], "value": "active"}},
                "properties": {
                    "name": {
                        "container": {"type": "container", "space": "my_space", "externalId": "my_container"},
                        "containerPropertyIdentifier": "name",
                    }
                },
            }
        },
    ]
    yield from (pytest.param(next(iter(f.values())), id=next(iter(f.keys()))) for f in view_filters)


class TestViewYAML:
    @pytest.mark.parametrize("data", list(find_resources("View")))
    def test_load_valid_view_file(self, data: dict[str, object]) -> None:
        loaded = ViewYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", view_property_type_cases())
    def test_load_valid_view_property_types(self, data: dict[str, object]) -> None:
        loaded = ViewYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", view_with_implements_cases())
    def test_load_valid_view_implements(self, data: dict[str, object]) -> None:
        loaded = ViewYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", view_with_filters_cases())
    def test_load_valid_view_filters(self, data: dict[str, object]) -> None:
        loaded = ViewYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    def test_invalid_external_id(self) -> None:
        invalid_data = {
            "space": "my_space",
            "externalId": "Query",  # Reserved name
            "version": "1",
            "properties": {
                "field": {
                    "container": {"type": "container", "space": "my_space", "externalId": "my_container"},
                    "containerPropertyIdentifier": "field",
                }
            },
        }

        with pytest.raises(ValueError, match="is a reserved view External ID"):
            ViewYAML.model_validate(invalid_data)

    def test_invalid_property_identifier(self) -> None:
        invalid_data = {
            "space": "my_space",
            "externalId": "valid_view",
            "version": "1",
            "properties": {
                "invalid@": {  # Invalid character
                    "container": {"type": "container", "space": "my_space", "externalId": "my_container"},
                    "containerPropertyIdentifier": "field",
                }
            },
        }

        with pytest.raises(ValueError, match="does not match the required pattern"):
            ViewYAML.model_validate(invalid_data)

    def test_forbidden_property_identifier(self) -> None:
        invalid_data = {
            "space": "my_space",
            "externalId": "valid_view",
            "version": "1",
            "properties": {
                "edge_id": {  # Reserved property name
                    "container": {"type": "container", "space": "my_space", "externalId": "my_container"},
                    "containerPropertyIdentifier": "field",
                }
            },
        }

        with pytest.raises(ValueError, match="is a reserved property identifier"):
            ViewYAML.model_validate(invalid_data)

    def test_space_format_validation(self) -> None:
        invalid_data = {
            "space": "invalid space",  # Contains a space
            "externalId": "valid_view",
            "version": "1",
            "properties": {
                "field": {
                    "container": {"type": "container", "space": "my_space", "externalId": "my_container"},
                    "containerPropertyIdentifier": "field",
                }
            },
        }

        with pytest.raises(ValueError):
            ViewYAML.model_validate(invalid_data)

    def test_invalid_version_format(self) -> None:
        invalid_data = {
            "space": "my_space",
            "externalId": "valid_view",
            "version": ".0.1",  # Invalid version format
            "properties": {
                "field": {
                    "container": {"type": "container", "space": "my_space", "externalId": "my_container"},
                    "containerPropertyIdentifier": "field",
                }
            },
        }

        with pytest.raises(ValueError):
            ViewYAML.model_validate(invalid_data)



import hashlib
import re
from collections.abc import Hashable

from rich.console import Console

from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.tk_warnings import LowSeverityWarning

INVALID_TITLE_REGEX = re.compile(r"[\\*?:/\[\]]")


def suffix_description(
    suffix: str,
    description: str | None,
    description_character_limit: int,
    identifier: Hashable,
    resource_type: str,
    console: Console | None = None,
) -> str:
    """Appends a suffix to a description if it is not already present.
    If the description is too long after appending the suffix, it will be truncated to fit within the character limit.

    Args:
        suffix: The suffix to append to the description.
        description: The original description to which the suffix will be appended.
        description_character_limit: The maximum number of characters allowed in the description after appending the suffix.
        identifier: Hashable identifier for the resource, used in warnings.
        resource_type:  Type of the resource, used in warnings.
        console: Console object for printing warnings.

    Returns:
        str: The modified description with the suffix appended, or truncated if necessary.
    """
    if description is None or description == "":
        return suffix
    elif description.endswith(suffix):
        # The suffix is already in the description
        return description
    elif len(description) + len(suffix) + 1 < description_character_limit:
        return f"{description} {suffix}"
    else:
        LowSeverityWarning(f"Description is too long for {resource_type} {identifier!r}. Truncating...").print_warning(
            console=console
        )
        truncation = description_character_limit - len(suffix) - 3
        return f"{description[:truncation]}...{suffix}"


def sanitize_spreadsheet_title(title: str) -> str:
    """Sanitizes a title for use in a spreadsheet by removing invalid characters.

    Args:
        title: The original title to sanitize.

    Returns:
        str: The sanitized title with invalid characters removed.
    """
    if not title:
        return "Sheet"
    sanitized_title = INVALID_TITLE_REGEX.sub("", title)
    return sanitized_title if sanitized_title else "Sheet"


def to_sentence_case(text: str) -> str:
    """Converts a string to sentence case.

    Args:
        text: The original string to convert.

    Returns:
        str: The string converted to sentence case.

    >>> to_sentence_case("hello world")
    'hello world'
    >>> to_sentence_case("PYTHON programming")
    'python programming'
    >>> to_sentence_case("")
    ''
    >>> to_sentence_case("snake_case_string")
    'snake case string'
    >>> to_sentence_case("MIXED_case STRING")
    'mixed case string'
    >>> to_sentence_case("camelCaseString")
    'camel case string'
    """
    if not text:
        return text
    # Replace underscores with spaces
    text = text.replace("_", " ")
    # Insert spaces before uppercase letters that are preceded by lowercase letters
    text = re.sub(r"(?<=[a-z])([A-Z])", r" \1", text)
    # Convert to lowercase
    return text.casefold()


def sanitize_instance_external_id(external_id: str) -> str:
    """Sanitize an instance external ID to be compatible with CDF requirements.

    Args:
        external_id: The external ID to sanitize.

    Returns:
        The sanitized external ID.
    """
    # CDF instance external IDs must be between 1 and 256 characters,
    if not external_id or external_id == "\x00":
        raise ToolkitValueError("External ID cannot be empty.")
    elif len(external_id) <= 256:
        return external_id
    hasher = hashlib.sha256()
    hasher.update(external_id.encode("utf-8"))
    hash_digest = hasher.hexdigest()[:8]
    sanitized_external_id = f"{external_id[:247]}_{hash_digest}"
    return sanitized_external_id



from abc import ABC, abstractmethod
from functools import cached_property
from typing import Annotated, Any, Literal

from cognite.client._proto.data_points_pb2 import (
    InstanceId,
)
from pydantic import Field

from ._base import DataSelector, SelectorObject


class Column(SelectorObject, ABC):
    column_type: str
    column: str
    dtype: Literal["numeric", "string"]

    @abstractmethod
    def as_wrapped_id(self) -> dict[str, Any]: ...


class InstanceColumn(Column):
    column_type: Literal["instance"] = "instance"
    space: str
    external_id: str

    def as_wrapped_id(self) -> dict[str, Any]:
        return {"instanceId": InstanceId(space=self.space, externalId=self.external_id)}


class ExternalIdColumn(Column):
    column_type: Literal["externalId"] = "externalId"
    external_id: str

    def as_wrapped_id(self) -> dict[str, Any]:
        return {"externalId": self.external_id}


class InternalIdColumn(Column):
    column_type: Literal["internalId"] = "internalId"
    internal_id: int

    def as_wrapped_id(self) -> dict[str, Any]:
        return {"id": self.internal_id}


TimeSeriesColumn = Annotated[
    InstanceColumn | ExternalIdColumn | InternalIdColumn,
    Field(discriminator="column_type"),
]


class DataPointsFileSelector(DataSelector):
    type: Literal["datapointsFile"] = "datapointsFile"
    kind: Literal["datapoints"] = "datapoints"

    timestamp_column: str
    columns: tuple[TimeSeriesColumn, ...]

    @property
    def group(self) -> str:
        return "Datapoints"

    def __str__(self) -> str:
        return "datapoints_file"

    @cached_property
    def id_by_column(self) -> dict[str, Column]:
        return {col.column: col for col in self.columns}



from cognite.client._api.raw import RawAPI

from cognite_toolkit._cdf_tk.client.data_classes.raw import RawProfileResults, RawTable


class ExtendedRawAPI(RawAPI):
    DEFAULT_PROFILE_LIMIT = 1000
    MAX_PROFILE_LIMIT = 1_000_000

    def profile(
        self,
        database: str | RawTable,
        table: str | None = None,
        limit: int = DEFAULT_PROFILE_LIMIT,
        timeout_seconds: int | None = None,
    ) -> RawProfileResults:
        """Profiles a table in the specified database and returns the results.

        This is a hidden endpoint that is not part of the official CDF API. However, it is used by the Fusion UI
        to profile tables in the database. This is implemented internally in Cognite Toolkit as Toolkit offers
        profiling of raw tables. This is used to show how data flows into CDF resources.

        Args:
            database (str): The name of the database to profile.
            table (str): The name of the table to profile.
            limit (int, optional): The maximum number of rows to profile. Defaults to DEFAULT_PROFILE_LIMIT.
            timeout_seconds (int, optional): The timeout for the profiling operation in seconds. Defaults to global_config.timeout_seconds.

        Returns:
            RawProfileResults: The results of the profiling operation.

        """
        if limit <= 0 or limit > self.MAX_PROFILE_LIMIT:
            raise ValueError(f"Limit must be between 1 and {self.MAX_PROFILE_LIMIT}, got {limit}.")
        if isinstance(database, RawTable):
            db_name = database.db_name
            table_name = database.table_name
        elif table is None:
            raise ValueError(f"Table name must be provided for {database}.")
        else:
            db_name = database
            table_name = table
        response = self._do_request(
            "POST",
            "/profiler/raw",
            json={"database": db_name, "table": table_name, "limit": limit},
            timeout=timeout_seconds if timeout_seconds is not None else self._config.timeout,
        )
        return RawProfileResults._load(response.json())



from collections.abc import Set
from typing import Any

import pytest
from cognite.client.data_classes.data_modeling import NodeId

from cognite_toolkit._cdf_tk.commands._migrate.data_classes import MigrationMapping, MigrationMappingList
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence, flatten_dict_json_path


class TestChunkSequence:
    def test_chunk_sequence(self):
        """Test that chunker_sequence correctly chunks a sequence."""
        sequence = list(range(10))
        chunk_size = 3
        expected_chunks = [
            [0, 1, 2],
            [3, 4, 5],
            [6, 7, 8],
            [9],
        ]

        chunks = list(chunker_sequence(sequence, chunk_size))
        assert chunks == expected_chunks

    def test_chunk_sequence_empty(self):
        """Test that chunker_sequence returns an empty list for an empty sequence."""
        sequence = []
        chunk_size = 3

        chunks = list(chunker_sequence(sequence, chunk_size))
        assert chunks == []

    def test_chunk_sequence_maintain_source_type(self) -> None:
        """Test that chunker_sequence maintains the type of the source sequence."""
        sequence = MigrationMappingList(
            [
                MigrationMapping(
                    resourceType="timeseries", id=1, dataSetId=123, instanceId=NodeId("sp_full_ts", "full_ts_id")
                ),
                MigrationMapping(
                    resourceType="timeseries", id=2, dataSetId=None, instanceId=NodeId("sp_step_ts", "step_ts_id")
                ),
            ]
        )
        chunk_size = 1

        chunks = list(chunker_sequence(sequence, chunk_size))
        assert len(chunks) == 2
        assert len(chunks[0]) == 1
        assert len(chunks[1]) == 1
        assert isinstance(chunks[0], MigrationMappingList)


class TestFlattenDictJsonPath:
    @pytest.mark.parametrize(
        "dct,exclude_keys,expected",
        [
            pytest.param({"a": 1}, set(), {"a": 1}, id="flat dict"),
            pytest.param({"a": {"b": 2}}, set(), {"a.b": 2}, id="nested dict"),
            pytest.param(
                {"a": {"b": [{"c": {"d": 3}}, 1]}}, set(), {"a.b[0].c.d": 3, "a.b[1]": 1}, id="nested dict with list"
            ),
            pytest.param({}, set(), {}, id="empty dict"),
            pytest.param({"a": []}, set(), {}, id="empty list"),
            pytest.param({"a": {}}, set(), {}, id="nested empty dict"),
            pytest.param({"a": None}, set(), {"a": None}, id="None value"),
            pytest.param({"a": [1, 2, 3]}, set(), {"a[0]": 1, "a[1]": 2, "a[2]": 3}, id="simple list"),
            pytest.param(
                {"a": [1, [2, 3], 4]},
                set(),
                {"a[0]": 1, "a[1][0]": 2, "a[1][1]": 3, "a[2]": 4},
                id="nested list",
            ),
            pytest.param(
                {"a": [1, [2, 3], 4]},
                {"a"},
                {"a": [1, [2, 3], 4]},
                id="Exclude list key",
            ),
            pytest.param(
                {"a": {"b": {"c": {"d": {"e": 5}}}}},
                set(),
                {"a.b.c.d.e": 5},
                id="deep nesting",
            ),
            pytest.param(
                {"a": {"b": [{"c": [1, {"d": 2}]}, 3]}},
                set(),
                {"a.b[0].c[0]": 1, "a.b[0].c[1].d": 2, "a.b[1]": 3},
                id="complex nesting",
            ),
            pytest.param(
                {"a": {"b": True, "c": False}},
                set(),
                {"a.b": True, "a.c": False},
                id="boolean values",
            ),
            pytest.param(
                {
                    "name": "example",
                    "labels": [{"externalId": "equipment"}, {"externalId": "Fornebu"}],
                    "metadata": {"createdBy": "user"},
                },
                {"labels", "metadata"},
                {
                    "name": "example",
                    "labels": [{"externalId": "equipment"}, {"externalId": "Fornebu"}],
                    "metadata": {"createdBy": "user"},
                },
                id="Exclude specific keys",
            ),
        ],
    )
    def test_flatten_dict_json_path(self, dct: dict[str, Any], exclude_keys: Set[str], expected: dict[str, Any]):
        """Test that flatten_dict_json_path correctly flattens a nested dictionary."""
        flat_dict = flatten_dict_json_path(dct, keep_structured=exclude_keys)
        assert flat_dict == expected



from __future__ import annotations

import os
import shutil
from collections.abc import Iterator
from pathlib import Path
from typing import Any

import pytest
import responses
import yaml
from cognite.client import global_config
from cognite.client.credentials import Token
from cognite.client.data_classes import CreatedSession
from cognite.client.data_classes.data_modeling import ContainerList, DataModel, NodeList, View, ViewId
from pytest import MonkeyPatch

from cognite_toolkit._cdf_tk.client import ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.canvas import IndustrialCanvas
from cognite_toolkit._cdf_tk.client.data_classes.migration import InstanceSource
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands import BuildCommand, ModulesCommand, RepoCommand
from cognite_toolkit._cdf_tk.constants import BUILD_ENVIRONMENT_FILE, MODULES
from cognite_toolkit._cdf_tk.data_classes._config_yaml import BuildEnvironment
from cognite_toolkit._cdf_tk.utils import read_yaml_file
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.constants import REPO_ROOT
from tests.data import (
    BUILDABLE_PACKAGE,
    COMPLETE_ORG,
    CORE_CONTAINERS_NO_3D_YAML,
    CORE_NO_3D_YAML,
    EXTRACTOR_VIEWS_YAML,
)
from tests.test_unit.approval_client import ApprovalToolkitClient
from tests.test_unit.utils import PrintCapture

THIS_FOLDER = Path(__file__).resolve().parent
TMP_FOLDER = THIS_FOLDER / "tmp"
TMP_FOLDER.mkdir(exist_ok=True)
BASE_URL = "http://blabla.cognitedata.com"


@pytest.fixture
def toolkit_client_approval() -> Iterator[ApprovalToolkitClient]:
    with monkeypatch_toolkit_client() as toolkit_client:

        def create_session(*args: Any, **kwargs: Any) -> CreatedSession:
            return CreatedSession(
                id=42,
                status="READY",
                nonce="dummy-nonce",
                type="CLIENT_CREDENTIALS",
            )

        toolkit_client.iam.sessions.create = create_session
        approval_client = ApprovalToolkitClient(toolkit_client)
        yield approval_client


@pytest.fixture(scope="function")
def env_vars_with_client(toolkit_client_approval: ApprovalToolkitClient) -> EnvironmentVariables:
    env_vars = EnvironmentVariables(
        CDF_CLUSTER="bluefield",
        CDF_PROJECT="pytest-project",
        LOGIN_FLOW="client_credentials",
        PROVIDER="entra_id",
        IDP_CLIENT_ID="dummy-123",
        IDP_CLIENT_SECRET="dummy-secret",
        IDP_TENANT_ID="dummy-domain",
    )
    env_vars._client = toolkit_client_approval.mock_client
    return env_vars


@pytest.fixture(scope="session")
def build_tmp_path() -> Iterator[Path]:
    pidid = os.getpid()
    build_folder = TMP_FOLDER / f"build-{pidid}"
    if build_folder.exists():
        shutil.rmtree(build_folder, ignore_errors=True)
        build_folder.mkdir(exist_ok=True)
    yield build_folder
    shutil.rmtree(build_folder, ignore_errors=True)


@pytest.fixture(scope="session")
def local_tmp_repo_path() -> Iterator[Path]:
    pidid = os.getpid()
    repo_path = TMP_FOLDER / f"pytest-repo-{pidid}"
    repo_path.mkdir(exist_ok=True)
    RepoCommand(silent=True, skip_git_verify=True).init(repo_path, host="GitHub")
    yield repo_path
    shutil.rmtree(repo_path, ignore_errors=True)


def init_organization_dir(organization_dir: Path, module_source_dir: Path) -> None:
    ModulesCommand(silent=True, module_source_dir=module_source_dir).init(
        organization_dir,
        select_all=True,
        clean=True,
    )


@pytest.fixture(scope="session")
def complete_org_dir(
    local_tmp_repo_path: Path,
) -> None:
    organization_dir = local_tmp_repo_path / "pytest-org"
    init_organization_dir(organization_dir, COMPLETE_ORG / MODULES)
    return organization_dir


@pytest.fixture(scope="session")
def complete_org_dir_mutable(
    local_tmp_repo_path: Path,
) -> Path:
    """This is used in tests were the source module files are modified. For example, cdf pull commands."""
    organization_dir = local_tmp_repo_path / "pytest-org-mutable"
    init_organization_dir(organization_dir, COMPLETE_ORG / MODULES)
    return organization_dir


@pytest.fixture(scope="session")
def buildable_modules(local_tmp_repo_path: Path) -> Path:
    organization_dir = local_tmp_repo_path / "legacy-pytest-org"
    init_organization_dir(organization_dir, BUILDABLE_PACKAGE / MODULES)
    return organization_dir


@pytest.fixture
def buildable_modules_mutable(
    local_tmp_repo_path: Path,
) -> Path:
    """This is used in tests were the source module files are modified. For example, cdf pull commands."""
    organization_dir = local_tmp_repo_path / "legacy-pytest-org-mutable"
    init_organization_dir(organization_dir, BUILDABLE_PACKAGE / MODULES)
    return organization_dir


@pytest.fixture
def build_environment(
    build_tmp_path: Path,
    complete_org_dir: Path,
    env_vars_with_client: EnvironmentVariables,
) -> BuildEnvironment:
    """Fixture that builds modules and returns the BuildEnvironment."""
    BuildCommand(silent=True, skip_tracking=True).execute(
        verbose=False,
        organization_dir=complete_org_dir,
        build_dir=build_tmp_path,
        selected=None,
        no_clean=False,
        client=env_vars_with_client.get_client(),
        build_env_name="dev",
        on_error="raise",
    )
    return BuildEnvironment.load(read_yaml_file(build_tmp_path / BUILD_ENVIRONMENT_FILE), "dev", "build")


@pytest.fixture
def capture_print(monkeypatch: MonkeyPatch) -> PrintCapture:
    capture = PrintCapture()
    toolkit_path = REPO_ROOT / "cognite_toolkit"
    monkeypatch.setattr("cognite_toolkit._cdf.print", capture)

    # Monkeypatch all print functions in the toolkit automatically
    for folder in ["_cdf_tk", "_api"]:
        for py_file in (toolkit_path / folder).rglob("*.py"):
            file_path = py_file.relative_to(toolkit_path)
            module_path = f"{'.'.join(['cognite_toolkit', *file_path.parts[:-1]])}.{file_path.stem}"
            try:
                monkeypatch.setattr(f"{module_path}.print", capture)
            except AttributeError:
                # The module does not have a print function
                continue
            except ImportError:
                raise

    return capture


@pytest.fixture
def disable_gzip():
    old = global_config.disable_gzip
    global_config.disable_gzip = True
    yield
    global_config.disable_gzip = old


@pytest.fixture
def disable_pypi_check():
    old = global_config.disable_pypi_version_check
    global_config.disable_pypi_version_check = True
    yield
    global_config.disable_pypi_version_check = old


@pytest.fixture
def toolkit_config():
    return ToolkitClientConfig(
        client_name="test-client",
        project="test-project",
        base_url=BASE_URL,
        max_workers=1,
        timeout=10,
        credentials=Token("abc"),
    )


@pytest.fixture
def rsps() -> Iterator[responses.RequestsMock]:
    with responses.RequestsMock() as rsps:
        yield rsps


@pytest.fixture(scope="session")
def asset_centric_canvas() -> tuple[IndustrialCanvas, NodeList[InstanceSource]]:
    canvas = IndustrialCanvas.load(
        {
            "annotations": [],
            "canvas": [
                {
                    "createdTime": 1751540227230,
                    "externalId": "495af88f-fe1d-403d-91b1-76ef9f80f265",
                    "instanceType": "node",
                    "lastUpdatedTime": 1751540558717,
                    "properties": {
                        "cdf_industrial_canvas": {
                            "Canvas/v7": {
                                "createdBy": "aGQ-cBXUBY6bmmxqIdkFoA",
                                "isArchived": None,
                                "isLocked": None,
                                "name": "Asset-centric1",
                                "solutionTags": None,
                                "sourceCanvasId": None,
                                "updatedAt": "2025-07-03T11:02:37.733+00:00",
                                "updatedBy": "aGQ-cBXUBY6bmmxqIdkFoA",
                                "visibility": "public",
                            }
                        }
                    },
                    "space": "IndustrialCanvasInstanceSpace",
                    "version": 14,
                }
            ],
            "containerReferences": [
                {
                    "createdTime": 1751540264906,
                    "externalId": "495af88f-fe1d-403d-91b1-76ef9f80f265_cf372b29-3012-49ff-8daf-5043404c23d7",
                    "instanceType": "node",
                    "lastUpdatedTime": 1751540264906,
                    "properties": {
                        "cdf_industrial_canvas": {
                            "ContainerReference/v2": {
                                "chartsId": None,
                                "containerReferenceType": "asset",
                                "height": 357,
                                "id": "cf372b29-3012-49ff-8daf-5043404c23d7",
                                "label": "Kelmarsh 6",
                                "maxHeight": None,
                                "maxWidth": None,
                                "resourceId": 3840956528416998,
                                "resourceSubId": None,
                                "width": 600,
                                "x": 0,
                                "y": 0,
                            }
                        }
                    },
                    "space": "IndustrialCanvasInstanceSpace",
                    "version": 1,
                },
                {
                    "createdTime": 1751540275336,
                    "externalId": "495af88f-fe1d-403d-91b1-76ef9f80f265_09d58ddf-bebb-4e4d-96db-1702da76a016",
                    "instanceType": "node",
                    "lastUpdatedTime": 1751540275336,
                    "properties": {
                        "cdf_industrial_canvas": {
                            "ContainerReference/v2": {
                                "chartsId": None,
                                "containerReferenceType": "timeseries",
                                "height": 400,
                                "id": "09d58ddf-bebb-4e4d-96db-1702da76a016",
                                "label": "Hub temperature, standard deviation (Â°C)",
                                "maxHeight": None,
                                "maxWidth": None,
                                "resourceId": 11978459264156,
                                "resourceSubId": None,
                                "width": 700,
                                "x": 700,
                                "y": 0,
                            }
                        }
                    },
                    "space": "IndustrialCanvasInstanceSpace",
                    "version": 1,
                },
                {
                    "createdTime": 1751540544349,
                    "externalId": "495af88f-fe1d-403d-91b1-76ef9f80f265_5e2bf845-103c-4c17-8549-9f20329b7f98",
                    "instanceType": "node",
                    "lastUpdatedTime": 1751540558717,
                    "properties": {
                        "cdf_industrial_canvas": {
                            "ContainerReference/v2": {
                                "chartsId": None,
                                "containerReferenceType": "event",
                                "height": 500,
                                "id": "5e2bf845-103c-4c17-8549-9f20329b7f98",
                                "label": "b18cdf8e-6568-4e2a-a267-535eb52f41bf",
                                "maxHeight": None,
                                "maxWidth": None,
                                "resourceId": 9004025980300864,
                                "resourceSubId": None,
                                "width": 600,
                                "x": -10,
                                "y": 418,
                            }
                        }
                    },
                    "space": "IndustrialCanvasInstanceSpace",
                    "version": 4,
                },
            ],
        }
    )
    mapping = NodeList[InstanceSource](
        [
            InstanceSource(
                space="MyNewInstanceSpace",
                external_id="my_asset",
                version=1,
                last_updated_time=1,
                created_time=1,
                resource_type="asset",
                id_=3840956528416998,
                preferred_consumer_view_id=ViewId("my_space", "DoctrinoAsset", "v1"),
            ),
            InstanceSource(
                space="MyNewInstanceSpace",
                external_id="my_timeseries",
                version=1,
                last_updated_time=1,
                created_time=1,
                resource_type="timeseries",
                id_=11978459264156,
                preferred_consumer_view_id=ViewId("my_space", "DoctrinoTimeSeries", "v1"),
            ),
            InstanceSource(
                space="MyNewInstanceSpace",
                external_id="my_event",
                version=1,
                last_updated_time=1,
                created_time=1,
                resource_type="event",
                id_=9004025980300864,
            ),
        ]
    )
    return canvas, mapping


@pytest.fixture(scope="session")
def cognite_core_no_3D() -> DataModel[View]:
    """This is a simplified CogniteCore data model without the 3D views.
    In addition, CogniteAsset does not implement CogniteVisualizable (which is also removed).

    Note if you use this fixture in a test, ensure that you do not modify the returned
    data model, as it is shared between tests.
    """
    return DataModel.load(CORE_NO_3D_YAML.read_text(encoding="utf-8"))


@pytest.fixture(scope="session")
def cognite_core_containers_no_3D() -> ContainerList:
    """This is a simplified list of containers from the cdf_cdm space without CogniteVisualizable.
    In addition, the CogniteAsset container does not require CogniteVisualizable.

    Note if you use this fixture in a test, ensure that you do not modify the returned
    containers, as it is shared between tests.
    """
    return ContainerList.load(CORE_CONTAINERS_NO_3D_YAML.read_text(encoding="utf-8"))


@pytest.fixture(scope="session")
def cognite_extractor_views() -> list[View]:
    """This is a simplified data model containing only the views used by the extractor.

    Note if you use this fixture in a test, ensure that you do not modify the returned
    data model, as it is shared between tests.
    """
    return [View._load(view) for view in yaml.safe_load(EXTRACTOR_VIEWS_YAML.read_text(encoding="utf-8"))]



from collections.abc import Mapping, Set
from dataclasses import dataclass
from typing import Any, ClassVar, overload

from cognite.client.data_classes import Annotation, Asset, Event, FileMetadata, TimeSeries
from cognite.client.data_classes.data_modeling import (
    DirectRelation,
    DirectRelationReference,
    EdgeId,
    MappedProperty,
    NodeApply,
    NodeId,
)
from cognite.client.data_classes.data_modeling.instances import EdgeApply, NodeOrEdgeData, PropertyValueWrite
from cognite.client.data_classes.data_modeling.views import ViewProperty

from cognite_toolkit._cdf_tk.client.data_classes.migration import AssetCentricId, ResourceViewMapping
from cognite_toolkit._cdf_tk.utils.collection import flatten_dict_json_path
from cognite_toolkit._cdf_tk.utils.dtype_conversion import (
    asset_centric_convert_to_primary_property,
    convert_to_primary_property,
)
from cognite_toolkit._cdf_tk.utils.useful_types import (
    AssetCentricResourceExtended,
    AssetCentricType,
)

from .data_model import INSTANCE_SOURCE_VIEW_ID
from .issues import ConversionIssue, FailedConversion, InvalidPropertyDataType


@dataclass
class DirectRelationCache:
    """Cache for direct relation references to look up target of direct relations.

    This is used when creating direct relations from asset-centric resources to CogniteAsset and CogniteSourceSystem.

    Attributes:
        asset: Mapping[int, DirectRelationReference]
            A mapping from asset IDs to their corresponding DirectRelationReference in the data model.
        source: Mapping[str, DirectRelationReference]
            A mapping from source strings to their corresponding DirectRelationReference in the data model.

    Methods:
        get(resource_type: AssetCentric, property_id: str) -> Mapping[str | int, DirectRelationReference]:
            Get the appropriate mapping based on the resource type and property ID.

    Note:
        The ASSET_REFERENCE_PROPERTIES and SOURCE_REFERENCE_PROPERTIES class variables define which properties
        in asset-centric resources reference assets and sources, respectively.

    """

    ASSET_REFERENCE_PROPERTIES: ClassVar[Set[tuple[AssetCentricType, str]]] = {
        ("timeseries", "assetId"),
        ("file", "assetIds"),
        ("event", "assetIds"),
        ("sequence", "assetId"),
        ("asset", "parentId"),
        ("fileAnnotation", "data.assetRef.id"),
    }
    SOURCE_REFERENCE_PROPERTIES: ClassVar[Set[tuple[AssetCentricType, str]]] = {
        ("asset", "source"),
        ("event", "source"),
        ("file", "source"),
    }
    FILE_REFERENCE_PROPERTIES: ClassVar[Set[tuple[AssetCentricType, str]]] = {
        ("fileAnnotation", "data.fileRef.id"),
        ("fileAnnotation", "annotatedResourceId"),
    }

    asset: Mapping[int, DirectRelationReference]
    source: Mapping[str, DirectRelationReference]
    file: Mapping[int, DirectRelationReference]

    def get(self, resource_type: AssetCentricType, property_id: str) -> Mapping[str | int, DirectRelationReference]:
        key = resource_type, property_id
        if key in self.ASSET_REFERENCE_PROPERTIES:
            return self.asset  # type: ignore[return-value]
        if key in self.SOURCE_REFERENCE_PROPERTIES:
            return self.source  # type: ignore[return-value]
        if key in self.FILE_REFERENCE_PROPERTIES:
            return self.file  # type: ignore[return-value]
        return {}


@overload
def asset_centric_to_dm(
    resource: AssetCentricResourceExtended,
    instance_id: NodeId,
    view_source: ResourceViewMapping,
    view_properties: dict[str, ViewProperty],
    asset_instance_id_by_id: Mapping[int, DirectRelationReference],
    source_instance_id_by_external_id: Mapping[str, DirectRelationReference],
    file_instance_id_by_id: Mapping[int, DirectRelationReference],
) -> tuple[NodeApply | None, ConversionIssue]: ...


@overload
def asset_centric_to_dm(
    resource: AssetCentricResourceExtended,
    instance_id: EdgeId,
    view_source: ResourceViewMapping,
    view_properties: dict[str, ViewProperty],
    asset_instance_id_by_id: Mapping[int, DirectRelationReference],
    source_instance_id_by_external_id: Mapping[str, DirectRelationReference],
    file_instance_id_by_id: Mapping[int, DirectRelationReference],
) -> tuple[EdgeApply | None, ConversionIssue]: ...


def asset_centric_to_dm(
    resource: AssetCentricResourceExtended,
    instance_id: NodeId | EdgeId,
    view_source: ResourceViewMapping,
    view_properties: dict[str, ViewProperty],
    asset_instance_id_by_id: Mapping[int, DirectRelationReference],
    source_instance_id_by_external_id: Mapping[str, DirectRelationReference],
    file_instance_id_by_id: Mapping[int, DirectRelationReference],
) -> tuple[NodeApply | EdgeApply | None, ConversionIssue]:
    """Convert an asset-centric resource to a data model instance.

    Args:
        resource (CogniteResource): The asset-centric resource to convert.
        instance_id (NodeId | EdgeApply): The ID of the instance to create or update.
        view_source (ResourceViewMapping): The view source defining how to map the resource to the data model.
        view_properties (dict[str, ViewProperty]): The defined properties referenced in the view source mapping.
        asset_instance_id_by_id (dict[int, DirectRelationReference]): A mapping from asset IDs to their corresponding
            DirectRelationReference in the data model. This is used to create direct relations for resources that
            reference assets.
        source_instance_id_by_external_id (dict[str, DirectRelationReference]): A mapping from source strings to their
            corresponding DirectRelationReference in the data model. This is used to create direct relations for resources
            that reference sources.
        file_instance_id_by_id (dict[int, DirectRelationReference]): A mapping from file IDs to their corresponding
            DirectRelationReference in the data model. This is used to create direct relations for resources that
            reference files.

    Returns:
        tuple[NodeApply | EdgeApply, ConversionIssue]: A tuple containing the converted NodeApply and any ConversionIssue encountered.
    """
    cache = DirectRelationCache(
        asset=asset_instance_id_by_id, source=source_instance_id_by_external_id, file=file_instance_id_by_id
    )
    resource_type = _lookup_resource_type(resource)
    dumped = resource.dump()
    try:
        id_ = dumped.pop("id")
    except KeyError as e:
        raise ValueError("Resource must have an 'id' field.") from e
    if not isinstance(id_, int):
        raise TypeError(f"Resource 'id' field must be an int, got {type(id_)}.")
    data_set_id = dumped.pop("dataSetId", None)
    external_id = dumped.pop("externalId", None)

    issue = ConversionIssue(asset_centric_id=AssetCentricId(resource_type, id_=id_), instance_id=instance_id)

    properties = create_properties(
        dumped,
        view_properties,
        view_source.property_mapping,
        resource_type,
        issue=issue,
        direct_relation_cache=cache,
    )
    sources: list[NodeOrEdgeData] = []
    if properties:
        sources.append(NodeOrEdgeData(source=view_source.view_id, properties=properties))

    if resource_type != "fileAnnotation":
        instance_source_properties = {
            "resourceType": resource_type,
            "id": id_,
            "dataSetId": data_set_id,
            "classicExternalId": external_id,
        }
        sources.append(NodeOrEdgeData(source=INSTANCE_SOURCE_VIEW_ID, properties=instance_source_properties))

    instance: NodeApply | EdgeApply
    if isinstance(instance_id, EdgeId):
        edge_properties = create_edge_properties(
            dumped, view_source.property_mapping, resource_type, issue, cache, instance_id.space
        )
        if any(key not in edge_properties for key in ("start_node", "end_node", "type")):
            # Failed conversion of edge properties
            return None, issue
        instance = EdgeApply(
            space=instance_id.space,
            external_id=instance_id.external_id,
            sources=sources,
            **edge_properties,  # type: ignore[arg-type]
        )
    elif isinstance(instance_id, NodeId):
        instance = NodeApply(space=instance_id.space, external_id=instance_id.external_id, sources=sources)
    else:
        raise RuntimeError(f"Unexpected instance_id type {type(instance_id)}")

    return instance, issue


def _lookup_resource_type(resource_type: AssetCentricResourceExtended) -> AssetCentricType:
    if isinstance(resource_type, Asset):
        return "asset"
    elif isinstance(resource_type, FileMetadata):
        return "file"
    elif isinstance(resource_type, Event):
        return "event"
    elif isinstance(resource_type, TimeSeries):
        return "timeseries"
    elif isinstance(resource_type, Annotation):
        if resource_type.annotated_resource_type == "file" and resource_type.annotation_type in (
            "diagrams.AssetLink",
            "diagrams.FileLink",
        ):
            return "fileAnnotation"
    raise ValueError(f"Unsupported resource type: {resource_type}")


def create_properties(
    dumped: dict[str, Any],
    view_properties: dict[str, ViewProperty],
    property_mapping: dict[str, str],
    resource_type: AssetCentricType,
    issue: ConversionIssue,
    direct_relation_cache: DirectRelationCache,
) -> dict[str, PropertyValueWrite]:
    """
    Create properties for a data model instance from an asset-centric resource.

    Args:
        dumped: Dict representation of the asset-centric resource.
        view_properties: Defined properties referenced in the view source mapping.
        property_mapping:  Mapping from asset-centric property IDs to data model property IDs.
        resource_type: The type of the asset-centric resource (e.g., "asset", "timeseries").
        issue: ConversionIssue object to log any issues encountered during conversion.
        direct_relation_cache: Cache for direct relation references to look up target of direct relations.

    Returns:
        Dict of property IDs to PropertyValueWrite objects.

    """
    flatten_dump = flatten_dict_json_path(dumped, keep_structured=set(property_mapping.keys()))
    properties: dict[str, PropertyValueWrite] = {}
    ignored_asset_centric_properties: set[str] = set()
    for prop_json_path, prop_id in property_mapping.items():
        if prop_json_path not in flatten_dump:
            continue
        if prop_id not in view_properties:
            continue
        if prop_id in properties:
            ignored_asset_centric_properties.add(prop_json_path)
            continue
        dm_prop = view_properties[prop_id]
        if not isinstance(dm_prop, MappedProperty):
            issue.invalid_instance_property_types.append(
                InvalidPropertyDataType(property_id=prop_id, expected_type=MappedProperty.__name__)
            )
            continue
        try:
            value = asset_centric_convert_to_primary_property(
                flatten_dump[prop_json_path],
                dm_prop.type,
                dm_prop.nullable,
                destination_container_property=(dm_prop.container, dm_prop.container_property_identifier),
                source_property=(resource_type, prop_json_path),
                direct_relation_lookup=direct_relation_cache.get(resource_type, prop_json_path),
            )
        except (ValueError, TypeError, NotImplementedError) as e:
            issue.failed_conversions.append(
                FailedConversion(property_id=prop_json_path, value=flatten_dump[prop_json_path], error=str(e))
            )
            continue
        properties[prop_id] = value

    issue.ignored_asset_centric_properties = sorted(
        (set(flatten_dump.keys()) - set(property_mapping.keys())) | ignored_asset_centric_properties
    )
    issue.missing_asset_centric_properties = sorted(set(property_mapping.keys()) - set(flatten_dump.keys()))
    # Node and edge properties are handled separately
    issue.missing_instance_properties = sorted(
        {
            prop_id
            for prop_id in property_mapping.values()
            if not (prop_id.startswith("edge.") or prop_id.startswith("node."))
        }
        - set(view_properties.keys())
    )
    return properties


def create_edge_properties(
    dumped: dict[str, Any],
    property_mapping: dict[str, str],
    resource_type: AssetCentricType,
    issue: ConversionIssue,
    direct_relation_cache: DirectRelationCache,
    default_instance_space: str,
) -> dict[str, DirectRelationReference]:
    flatten_dump = flatten_dict_json_path(dumped)
    edge_properties: dict[str, DirectRelationReference] = {}
    for prop_json_path, prop_id in property_mapping.items():
        if not prop_id.startswith("edge."):
            continue
        if prop_json_path not in flatten_dump:
            continue
        edge_prop_id = prop_id.removeprefix("edge.")
        if edge_prop_id in ("startNode", "endNode", "type"):
            # DirectRelation lookup.
            try:
                value = convert_to_primary_property(
                    flatten_dump[prop_json_path],
                    DirectRelation(),
                    False,
                    direct_relation_lookup=direct_relation_cache.get(resource_type, prop_json_path),
                )
            except (ValueError, TypeError, NotImplementedError) as e:
                issue.failed_conversions.append(
                    FailedConversion(property_id=prop_json_path, value=flatten_dump[prop_json_path], error=str(e))
                )
                continue
        elif edge_prop_id.endswith(".externalId"):
            # Just an external ID string.
            edge_prop_id = edge_prop_id.removesuffix(".externalId")
            value = DirectRelationReference(default_instance_space, str(flatten_dump[prop_json_path]))
        else:
            issue.invalid_instance_property_types.append(
                InvalidPropertyDataType(property_id=prop_id, expected_type="EdgeProperty")
            )
            continue
        # We know that value is DirectRelationReference here
        edge_properties[edge_prop_id.replace("Node", "_node")] = value  # type: ignore[assignment]

    return edge_properties



import gzip
import time
from collections.abc import MutableMapping, Sequence
from typing import Any

from cognite.client import global_config
from cognite.client._api.data_modeling import DataModelingAPI
from cognite.client._api.data_modeling.instances import InstancesAPI
from cognite.client._api_client import T
from cognite.client._cognite_client import ClientConfig, CogniteClient
from cognite.client._http_client import HTTPClient, HTTPClientConfig, get_global_requests_session
from cognite.client.data_classes.data_modeling import EdgeId, InstanceApply, NodeId
from cognite.client.exceptions import CogniteConnectionError, CogniteReadTimeout
from cognite.client.utils import _json
from cognite.client.utils._concurrency import execute_tasks
from requests import Response

from cognite_toolkit._cdf_tk.client._constants import DATA_MODELING_MAX_DELETE_WORKERS, DATA_MODELING_MAX_WRITE_WORKERS
from cognite_toolkit._cdf_tk.client.data_classes.instances import InstancesApplyResultList
from cognite_toolkit._cdf_tk.client.utils._concurrency import ToolkitConcurrencySettings
from cognite_toolkit._cdf_tk.client.utils._http_client import ToolkitRetryTracker


class ExtendedDataModelingAPI(DataModelingAPI):
    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self.instances: ExtendedInstancesAPI = ExtendedInstancesAPI(config, api_version, cognite_client)


class ExtendedInstancesAPI(InstancesAPI):
    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        # The self._http_client retries 429 status codes, which is not desired when we want to reduce the number of items
        # sent in each request when a 400 error occurs.
        self._http_client_no_retry = HTTPClient(
            config=HTTPClientConfig(
                status_codes_to_retry=set(),
                backoff_factor=0.5,
                max_backoff_seconds=global_config.max_retry_backoff,
                max_retries_total=0,
                max_retries_read=0,
                max_retries_connect=0,
                max_retries_status=0,
            ),
            session=get_global_requests_session(),
            refresh_auth_header=self._refresh_auth_header,
        )

    def apply_fast(
        self,
        items: Sequence[InstanceApply],
        auto_create_start_nodes: bool = False,
        auto_create_end_nodes: bool = False,
        auto_create_direct_relations: bool = True,
        skip_on_version_conflict: bool = False,
        replace: bool = False,
    ) -> InstancesApplyResultList:
        """Add or update (upsert) instances. <https://developer.cognite.com/api#tag/Instances/operation/applyNodeAndEdges>`_

        This is an alternative to the standard apply method that is available in the CDF Python SDK. The .apply
        method has two limitations that this method does not have:
        1. The standard apply method uses a thread pool executor with only one worker, even though the API supports 4.
           This method allows up to 4 workers to be used, which can significantly speed up ingestion of large datasets.
        2. The standard apply method has a generic handling of 400 errors. This method dynamically reduces the number
           of items sent in each request if a 400 error occurs.

        Args:
            items: A sequence of instances to apply. Each item can be a NodeApply or an EdgeApply instance.
            auto_create_start_nodes (bool): Whether to create missing start nodes for edges when ingesting. By default, the start node of an edge must exist before it can be ingested.
            auto_create_end_nodes (bool): Whether to create missing end nodes for edges when ingesting. By default, the end node of an edge must exist before it can be ingested.
            auto_create_direct_relations (bool): Whether to create missing direct relation targets when ingesting.
            skip_on_version_conflict (bool): If existingVersion is specified on any of the nodes/edges in the input, the default behaviour is that the entire ingestion will fail when version conflicts occur. If skipOnVersionConflict is set to true, items with version conflicts will be skipped instead. If no version is specified for nodes/edges, it will do the writing directly.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)? Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.

        Returns:
            InstancesApplyResult: Created instance(s)

        """
        parameters = {
            "autoCreateStartNodes": auto_create_start_nodes,
            "autoCreateEndNodes": auto_create_end_nodes,
            "autoCreateDirectRelations": auto_create_direct_relations,
            "skipOnVersionConflict": skip_on_version_conflict,
            "replace": replace,
        }
        tasks = [
            (self._RESOURCE_PATH, task_items, ToolkitRetryTracker(self._http_client_with_retry.config))
            for task_items in self._prepare_item_chunks(items, self._CREATE_LIMIT, parameters)
        ]
        summary = execute_tasks(
            self._post_with_item_reduction_retry,
            tasks,
            max_workers=min(self._config.max_workers, DATA_MODELING_MAX_WRITE_WORKERS),
            executor=ToolkitConcurrencySettings.get_data_modeling_write_executor(),
        )

        def unwrap_element(el: T) -> InstanceApply | NodeId | EdgeId | T:
            if isinstance(el, dict):
                instance_type = el.get("instanceType")
                if instance_type == "node":
                    return NodeId.load(el)
                elif instance_type == "edge":
                    return EdgeId.load(el)
                else:
                    return InstanceApply._load(el)
            else:
                return el

        summary.raise_compound_exception_if_failed_tasks(
            task_unwrap_fn=lambda task: task[1]["items"],
            task_list_element_unwrap_fn=unwrap_element,
        )
        created_resources = summary.joined_results(lambda res: res.json()["items"])

        return InstancesApplyResultList._load(created_resources)

    def delete_fast(
        self,
        instance_ids: Sequence[NodeId | EdgeId],
    ) -> list[NodeId | EdgeId]:
        """`Delete one or more instances <https://developer.cognite.com/api#tag/Instances/operation/deleteBulk>`_

        Args:
            instance_ids (Sequence[NodeId | EdgeId]): A sequence of NodeId or EdgeId instances to delete.

        Returns:
            list[NodeId | EdgeId]: A list of NodeId or EdgeId instances that were successfully deleted. An empty
                list is returned if no instances were deleted.
        """
        tasks = [
            (f"{self._RESOURCE_PATH}/delete", task_items, ToolkitRetryTracker(self._http_client_with_retry.config))
            for task_items in self._prepare_item_chunks(
                [
                    {"space": item.space, "externalId": item.external_id, "instanceType": item._instance_type}
                    for item in instance_ids
                ],
                self._DELETE_LIMIT,
                None,
            )
        ]
        summary = execute_tasks(
            self._post_with_item_reduction_retry,
            tasks,
            max_workers=min(self._config.max_workers, DATA_MODELING_MAX_DELETE_WORKERS),
            executor=ToolkitConcurrencySettings.get_data_modeling_delete_executor(),
        )

        def unwrap_element(el: T) -> InstanceApply | T:
            if isinstance(el, dict) and "instanceType" in el:
                if el["instanceType"] == "node":
                    return NodeId.load(el)
                elif el["instanceType"] == "edge":
                    return EdgeId.load(el)
                return el
            else:
                return el

        summary.raise_compound_exception_if_failed_tasks(
            task_unwrap_fn=lambda task: task[1]["items"],
            task_list_element_unwrap_fn=unwrap_element,
        )
        deleted_resources = summary.joined_results(lambda res: res.json()["items"])
        result: list[NodeId | EdgeId] = []
        for resource in deleted_resources:
            if "instanceType" not in resource:
                raise ValueError("Resource must contain 'instanceType' key.")
            instance_type = resource.get("instanceType")
            if instance_type == "node":
                result.append(NodeId.load(resource))
            elif instance_type == "edge":
                result.append(EdgeId.load(resource))
            else:
                raise TypeError(f"Resource must be a NodeId or EdgeId, not {instance_type}.")
        return result

    def _post_with_item_reduction_retry(
        self,
        url_path: str,
        json_payload: dict[str, Any],
        retry_tracker: ToolkitRetryTracker,
    ) -> Response:
        _, full_url = self._resolve_url("POST", url_path)
        headers = self._configure_headers(
            "application/json",
            additional_headers=self._config.headers.copy(),
        )
        data: str | bytes
        try:
            data = _json.dumps(json_payload, allow_nan=False)
        except ValueError as e:
            # A lot of work to give a more human friendly error message when nans and infs are present:
            msg = "Out of range float values are not JSON compliant"
            if msg in str(e):  # exc. might e.g. contain an extra ": nan", depending on build (_json.make_encoder)
                raise ValueError(f"{msg}. Make sure your data does not contain NaN(s) or +/- Inf!").with_traceback(
                    e.__traceback__
                ) from None
            raise

        if not global_config.disable_gzip:
            data = gzip.compress(data.encode())
            headers["Content-Encoding"] = "gzip"

        try:
            result = self._http_client_no_retry.request(
                "POST", full_url, data=data, headers=headers, timeout=self._config.timeout
            )
        except CogniteReadTimeout as e:
            retry_tracker.read += 1
            if not retry_tracker.should_retry(status_code=None, is_auto_retryable=True):
                raise e
            self._sleep(retry_tracker, headers)
            return self._post_with_item_reduction_retry(url_path, json_payload, retry_tracker)
        except CogniteConnectionError as e:
            retry_tracker.connect += 1
            if not retry_tracker.should_retry(status_code=None, is_auto_retryable=True):
                raise e
            self._sleep(retry_tracker, headers)
            return self._post_with_item_reduction_retry(url_path, json_payload, retry_tracker)

        retry_tracker.status += 1
        match result.status_code:
            case 200 | 201 | 202 | 204:
                pass
            case 401:
                self._raise_no_project_access_error(result)
            case 429 | 502 | 503 | 504:
                items = json_payload["items"]
                if retry_tracker.should_retry(status_code=result.status_code, is_auto_retryable=True):
                    if len(items) == 1:
                        self._sleep(retry_tracker, headers)
                        return self._post_with_item_reduction_retry(url_path, json_payload, retry_tracker)
                    # If we get a 429 or 5xx error, we reduce the number of items in the payload and retry.
                    # This is to avoid overwhelming the API with too many items at once.
                    half = len(items) // 2
                    first, second = json_payload.copy(), json_payload.copy()
                    first["items"] = items[:half]
                    second["items"] = items[half:]
                    self._sleep(retry_tracker, headers)
                    first_result = self._post_with_item_reduction_retry(url_path, first, retry_tracker.copy())
                    second_result = self._post_with_item_reduction_retry(url_path, second, retry_tracker.copy())
                    first_result.json()["items"].extend(second_result.json()["items"])
                    return first_result
                self._raise_api_error(result, payload=json_payload)
            case _:
                self._raise_api_error(result, payload=json_payload)

        self._log_request(result, payload=json_payload)
        return result

    def _sleep(self, retry_tracker: ToolkitRetryTracker, headers: MutableMapping[str, Any]) -> None:
        # During a backoff loop, our credentials might expire, so we check and maybe refresh:
        time.sleep(retry_tracker.get_backoff_time())
        self._refresh_auth_header(headers)



from collections.abc import Iterable, Sequence
from typing import Any, overload

from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.data_modeling import (
    EdgeId,
    InstanceSort,
    NodeApplyResultList,
    NodeId,
    NodeList,
    filters,
    query,
)
from cognite.client.data_classes.filters import Filter
from cognite.client.exceptions import CogniteDuplicatedError
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.canvas import (
    ANNOTATION_EDGE_TYPE,
    CANVAS_INSTANCE_SPACE,
    CONTAINER_REFERENCE_EDGE_TYPE,
    FDM_CONTAINER_REFERENCE_EDGE_TYPE,
    Canvas,
    CanvasAnnotation,
    CanvasApply,
    CogniteSolutionTag,
    ContainerReference,
    FdmInstanceContainerReference,
    IndustrialCanvas,
    IndustrialCanvasApply,
)
from cognite_toolkit._cdf_tk.client.data_classes.instances import InstancesApplyResultList
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError

from .extended_data_modeling import ExtendedInstancesAPI


class CanvasAPI:
    def __init__(self, instance_api: ExtendedInstancesAPI) -> None:
        self._instance_api = instance_api
        self.instance_space = CANVAS_INSTANCE_SPACE
        self.industrial = IndustrialCanvasAPI(instance_api)

    def upsert(self, canvas: CanvasApply | Sequence[CanvasApply]) -> NodeApplyResultList:
        return self._instance_api.apply(canvas).nodes

    def delete(self, external_id: str | SequenceNotStr[str]) -> list[str]:
        external_ids = [external_id] if isinstance(external_id, str) else external_id
        result = self._instance_api.delete([NodeId(self.instance_space, item) for item in external_ids])
        return [item.external_id for item in result.nodes]

    @overload
    def retrieve(self, external_id: str | None = None) -> Canvas | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str]) -> NodeList[Canvas]: ...

    def retrieve(self, external_id: str | SequenceNotStr[str] | None = None) -> Canvas | NodeList[Canvas] | None:
        if isinstance(external_id, str):
            return self._instance_api.retrieve_nodes(NodeId(self.instance_space, external_id), node_cls=Canvas)
        elif isinstance(external_id, Iterable):
            return self._instance_api.retrieve_nodes(
                [NodeId(self.instance_space, item) for item in external_id], node_cls=Canvas
            )
        else:
            raise TypeError(f"Expected str or SequenceNotStr[str], got {type(external_id)}")

    def list(
        self,
        limit: int | None = DEFAULT_LIMIT_READ,
        sort: Sequence[InstanceSort] | InstanceSort | None = None,
        filter: Filter | dict[str, Any] | None = None,
    ) -> NodeList[Canvas]:
        return self._instance_api.list(
            instance_type=Canvas, space=self.instance_space, limit=limit, sort=sort, filter=filter
        )


class IndustrialCanvasAPI:
    def __init__(self, instance_api: ExtendedInstancesAPI) -> None:
        self._instance_api = instance_api
        self._APPLY_LIMIT = 1000  # Limit for applying instances in a single request

    def create(self, canvas: IndustrialCanvasApply) -> InstancesApplyResultList:
        instance_ids = canvas.as_instance_ids()
        existing = self._instance_api.retrieve(
            [node for node in instance_ids if isinstance(node, NodeId)],
            edges=[edge for edge in instance_ids if isinstance(edge, EdgeId)],
        )
        if existing.nodes or existing.edges:
            raise CogniteDuplicatedError(duplicated=existing.nodes.as_ids() + existing.edges.as_ids())
        instances = canvas.as_instances()
        self._validate_instance_count(len(instances))
        return self._instance_api.apply_fast(instances)

    def retrieve(self, external_id: str) -> IndustrialCanvas | None:
        retrieve_query = self._retrieve_query(external_id)
        result = self._instance_api.query(retrieve_query)
        if len(result["canvas"]) == 0:
            return None
        # MyPy fails to understand that query result is a dict.
        return IndustrialCanvas._load(result)  # type: ignore[arg-type]

    def update(self, canvas: IndustrialCanvasApply) -> InstancesApplyResultList:
        new_instance_ids = canvas.as_instance_ids(include_solution_tags=False)
        self._validate_instance_count(len(new_instance_ids))

        existing = self.retrieve(external_id=canvas.as_id())
        if existing is None:
            raise ToolkitValueError(
                f"Cannot update canvas. Industrial canvas with external ID '{canvas.as_id()}' does not exist."
            )
        existing_instance_ids = existing.as_write().as_instance_ids(include_solution_tags=False)
        to_delete = set(existing_instance_ids) - set(new_instance_ids)
        result = self._instance_api.apply_fast(canvas.as_instances())
        if to_delete:
            # Delete components that are not in the new canvas
            self._instance_api.delete_fast(list(to_delete))
        return result

    def delete(self, canvas: IndustrialCanvasApply) -> list[NodeId | EdgeId]:
        # Solution tags are used by multiple canvases, so we do not include them in the deletion.
        return self._instance_api.delete_fast(canvas.as_instance_ids(include_solution_tags=False))

    def _validate_instance_count(self, instance_count: int) -> None:
        if instance_count > self._APPLY_LIMIT:
            # Looking at the largest Canvas users I have not seen any larger than 500 components (node + edge)
            # Creating a canvas with more than 1000 instances is more complex as it is not an atomic operation
            # thus postponing that implementation for now.
            raise ToolkitValueError(
                f"Creating/Updating an industrial canvas with more than {self._APPLY_LIMIT // 2} components is not supported. "
                "Please contact support if you need to create a larger canvas."
            )

    @classmethod
    def _retrieve_query(cls, external_id: str) -> query.Query:
        return query.Query(
            with_={
                "canvas": query.NodeResultSetExpression(
                    filter=filters.InstanceReferences([(CANVAS_INSTANCE_SPACE, external_id)]),
                    limit=1,
                ),
                "solutionTags": query.NodeResultSetExpression(
                    from_="canvas",
                    through=Canvas.get_source().as_property_ref("solutionTags"),
                ),
                "annotationEdges": query.EdgeResultSetExpression(
                    from_="canvas",
                    filter=filters.Equals(["edge", "type"], ANNOTATION_EDGE_TYPE.dump()),
                    node_filter=filters.HasData(views=[CanvasAnnotation.get_source()]),
                    direction="outwards",
                ),
                "containerReferenceEdges": query.EdgeResultSetExpression(
                    from_="canvas",
                    filter=filters.Equals(["edge", "type"], CONTAINER_REFERENCE_EDGE_TYPE.dump()),
                    node_filter=filters.HasData(views=[ContainerReference.get_source()]),
                    direction="outwards",
                ),
                "fdmInstanceContainerReferenceEdges": query.EdgeResultSetExpression(
                    from_="canvas",
                    filter=filters.Equals(
                        ["edge", "type"],
                        FDM_CONTAINER_REFERENCE_EDGE_TYPE.dump(),
                    ),
                    node_filter=filters.HasData(views=[FdmInstanceContainerReference.get_source()]),
                    direction="outwards",
                ),
                "annotations": query.NodeResultSetExpression(from_="annotationEdges"),
                "containerReferences": query.NodeResultSetExpression(from_="containerReferenceEdges"),
                "fdmInstanceContainerReferences": query.NodeResultSetExpression(
                    from_="fdmInstanceContainerReferenceEdges"
                ),
            },
            select={
                "canvas": query.Select([query.SourceSelector(Canvas.get_source(), properties=["*"])]),
                "solutionTags": query.Select([query.SourceSelector(CogniteSolutionTag.get_source(), properties=["*"])]),
                "annotations": query.Select([query.SourceSelector(CanvasAnnotation.get_source(), properties=["*"])]),
                "containerReferences": query.Select(
                    [query.SourceSelector(ContainerReference.get_source(), properties=["*"])]
                ),
                "fdmInstanceContainerReferences": query.Select(
                    [query.SourceSelector(FdmInstanceContainerReference.get_source(), properties=["*"])]
                ),
            },
        )



import json
from pathlib import Path
from unittest.mock import MagicMock

import httpx
import pytest
import respx
from cognite.client.data_classes import (
    Asset,
    AssetList,
    CountAggregate,
    DataSet,
    DataSetList,
    Event,
    EventList,
    FileMetadata,
    FileMetadataList,
    LabelDefinition,
    LabelDefinitionList,
    TimeSeries,
    TimeSeriesList,
)

from cognite_toolkit._cdf_tk.client import ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands import DownloadCommand, UploadCommand
from cognite_toolkit._cdf_tk.storageio import AssetIO, EventIO, FileMetadataIO, TimeSeriesIO
from cognite_toolkit._cdf_tk.storageio.selectors import AssetSubtreeSelector, DataSetSelector
from cognite_toolkit._cdf_tk.utils.collection import chunker
from cognite_toolkit._cdf_tk.utils.fileio import FileReader
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal


@pytest.fixture()
def some_asset_data() -> AssetList:
    """Fixture to provide a sample AssetList for testing."""
    return AssetList(
        [
            Asset(
                id=1000 + i,
                external_id=f"asset_{i}",
                name=f"Asset {i}",
                description=f"Description for asset {i}",
                root_id=123,
                source="test_source",
                labels=["my_label"],
                data_set_id=1234,
            )
            for i in range(100)
        ]
    )


class TestAssetIO:
    @pytest.mark.usefixtures("disable_gzip", "disable_pypi_check")
    def test_download_upload(
        self, toolkit_config: ToolkitClientConfig, some_asset_data: AssetList, respx_mock: respx.MockRouter
    ) -> None:
        config = toolkit_config
        asset_by_external_id = {asset.external_id: asset for asset in some_asset_data if asset.external_id is not None}
        selector = AssetSubtreeSelector(hierarchy="test_hierarchy", kind="Assets")

        def create_callback(request: httpx.Request) -> httpx.Response:
            payload = json.loads(request.content)
            assert "items" in payload
            items = payload["items"]
            assert isinstance(items, list)
            return httpx.Response(
                status_code=200,
                json={
                    "items": [asset_by_external_id[item["externalId"]].dump() for item in items if "externalId" in item]
                },
            )

        respx_mock.post(config.create_api_url("/assets")).mock(side_effect=create_callback)
        with monkeypatch_toolkit_client() as client:
            client.config = config
            client.assets.return_value = chunker(some_asset_data, 10)
            client.assets.aggregate_count.return_value = 100
            client.lookup.data_sets.external_id.return_value = "test_data_set"
            client.lookup.data_sets.id.return_value = 1234

            io = AssetIO(client)

            assert io.count(selector) == 100

            source = io.stream_data(selector)
            json_chunks: list[list[dict[str, JsonVal]]] = []
            for page in source:
                # New interface: stream_data returns Page objects
                json_chunk = io.data_to_json_chunk(page.items)
                assert isinstance(json_chunk, list)
                assert len(json_chunk) == 10
                for item in json_chunk:
                    assert isinstance(item, dict)
                    assert "dataSetExternalId" in item
                    assert item["dataSetExternalId"] == "test_data_set"
                json_chunks.append(json_chunk)

            with HTTPClient(config) as upload_client:
                from cognite_toolkit._cdf_tk.storageio._base import UploadItem

                # New interface: convert individual items and create UploadItems
                for chunk in json_chunks:
                    write_items = [io.json_to_resource(item) for item in chunk]
                    upload_items = [UploadItem(source_id=io.as_id(item), item=item) for item in write_items]
                    io.upload_items(upload_items, upload_client, selector)

            assert respx_mock.calls.call_count == 10  # 100 rows in chunks of 10
            uploaded_assets = []
            for call in respx_mock.calls:
                uploaded_assets.extend(json.loads(call.request.content)["items"])

            assert uploaded_assets == some_asset_data.as_write().dump()

    @pytest.mark.usefixtures("disable_gzip", "disable_pypi_check")
    def test_download_upload_command(
        self,
        some_asset_data: AssetList,
        tmp_path: Path,
        toolkit_config: ToolkitClientConfig,
        respx_mock: respx.MockRouter,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        config = toolkit_config
        monkeypatch.setenv("CDF_CLUSTER", config.cdf_cluster)
        monkeypatch.setenv("CDF_PROJECT", config.project)

        def asset_create_callback(request: httpx.Request) -> httpx.Response:
            payload = json.loads(request.content)
            assert "items" in payload
            items = payload["items"]
            assert isinstance(items, list)
            assert items == [asset.as_write().dump() for asset in some_asset_data]
            return httpx.Response(status_code=200, json={"items": some_asset_data.dump()})

        respx_mock.post(config.create_api_url("/assets")).mock(side_effect=asset_create_callback)

        selector = AssetSubtreeSelector(hierarchy="test_hierarchy", kind="Assets")
        with monkeypatch_toolkit_client() as client:
            client.config = config
            client.verify.authorization.return_value = []
            client.assets.return_value = [some_asset_data]
            client.assets.aggregate_count.return_value = 100
            client.lookup.data_sets.external_id.return_value = "test_data_set"
            client.lookup.data_sets.id.return_value = 1234
            client.lookup.assets.external_id.return_value = "test_hierarchy"
            client.lookup.assets.id.return_value = 123
            client.data_sets.retrieve_multiple.return_value = DataSetList(
                [DataSet(id=1234, external_id="test_data_set")]
            )
            client.labels.retrieve.return_value = LabelDefinitionList(
                [LabelDefinition(external_id="my_label", name="my_label")]
            )

            io = AssetIO(client)

            download_command = DownloadCommand(silent=True, skip_tracking=True)
            upload_command = UploadCommand(silent=True, skip_tracking=True)

            download_command.download(
                selectors=[selector],
                io=io,
                output_dir=tmp_path,
                verbose=False,
                file_format=".csv",
                compression="none",
                limit=100,
            )

            upload_command.upload(
                input_dir=tmp_path / selector.group,
                client=client,
                deploy_resources=True,
                dry_run=False,
                verbose=False,
                kind=io.KIND,
            )

            assert len(respx_mock.calls) == 1

    def test_read_assets_chunks(self) -> None:
        assets = [
            {"id": 1, "depth": 3},
            {"id": 2, "depth": 2},
            {"id": 3, "depth": 1},
            {"id": 4},
            {"id": 5, "depth": "not_an_int"},
        ]
        assets_with_line_numbers = list(enumerate(assets, start=1))
        other_reader = MagicMock(spec=FileReader)
        other_reader.read_chunks_with_line_numbers.return_value = assets_with_line_numbers
        other_reader.input_file = Path("mocked_file.csv")
        output = list(AssetIO.read_chunks(other_reader))

        assert output == [
            [("line 4", {"id": 4}), ("line 5", {"id": 5, "depth": "not_an_int"})],
            [("line 3", {"id": 3, "depth": 1})],
            [("line 2", {"id": 2, "depth": 2})],
            [("line 1", {"id": 1, "depth": 3})],
        ]


@pytest.fixture()
def some_filemetadata_data() -> FileMetadataList:
    """Fixture to provide a sample FileMetadataList for testing."""
    return FileMetadataList(
        [
            FileMetadata(
                external_id=f"file_{i}",
                name=f"File {i}",
                directory="/test/dir",
                mime_type="text/plain",
                data_set_id=1234,
                asset_ids=[123],
                source="test_source",
            )
            for i in range(50)
        ]
    )


class TestFileMetadataIO:
    @pytest.mark.usefixtures("disable_gzip", "disable_pypi_check")
    def test_download_upload(
        self,
        toolkit_config: ToolkitClientConfig,
        some_filemetadata_data: FileMetadataList,
        respx_mock: respx.MockRouter,
    ) -> None:
        config = toolkit_config
        file_by_external_id = {
            file.external_id: file for file in some_filemetadata_data if file.external_id is not None
        }

        def create_callback(request: httpx.Request) -> httpx.Response:
            payload = json.loads(request.content)
            assert "externalId" in payload
            return httpx.Response(
                status_code=200,
                json={"items": [file_by_external_id[payload["externalId"]].dump()]},
            )

        respx_mock.post(config.create_api_url("/files")).mock(side_effect=create_callback)
        selector = DataSetSelector(data_set_external_id="DataSetSelector", kind="FileMetadata")

        with monkeypatch_toolkit_client() as client:
            client.config = config
            client.files.return_value = chunker(some_filemetadata_data, 10)
            client.files.aggregate.return_value = [CountAggregate(50)]
            client.lookup.data_sets.external_id.return_value = "test_data_set"
            client.lookup.data_sets.id.return_value = 1234
            client.lookup.assets.external_id.return_value = ["test_hierarchy"]
            client.lookup.assets.id.return_value = [123]

            io = FileMetadataIO(client)

            assert io.count(selector) == 50

            source = io.stream_data(selector)
            json_chunks: list[list[dict[str, JsonVal]]] = []
            for page in source:
                # New interface: stream_data returns Page objects
                json_chunk = io.data_to_json_chunk(page.items)
                assert isinstance(json_chunk, list)
                assert len(json_chunk) == 10
                for item in json_chunk:
                    assert isinstance(item, dict)
                    assert "dataSetExternalId" in item
                    assert item["dataSetExternalId"] == "test_data_set"
                json_chunks.append(json_chunk)

            with HTTPClient(config) as upload_client:
                from cognite_toolkit._cdf_tk.storageio._base import UploadItem

                # New interface: convert individual items and create UploadItems
                for chunk in json_chunks:
                    write_items = [io.json_to_resource(item) for item in chunk]
                    upload_items = [UploadItem(source_id=io.as_id(item), item=item) for item in write_items]
                    io.upload_items(upload_items, upload_client, selector)

            # /files only support creating one at a time.
            assert respx_mock.calls.call_count == len(some_filemetadata_data)
            uploaded_files = []
            for call in respx_mock.calls:
                uploaded_files.append(json.loads(call.request.content))

            assert uploaded_files == some_filemetadata_data.as_write().dump()


@pytest.fixture()
def some_timeseries_data() -> TimeSeriesList:
    """Fixture to provide a sample TimeSeriesList for testing."""
    return TimeSeriesList(
        [
            TimeSeries(
                external_id=f"ts_{i}",
                name=f"Time Series {i}",
                description=f"Description for time series {i}",
                asset_id=123,
                data_set_id=1234,
                unit="unit",
                is_string=False,
                is_step=False,
            )
            for i in range(50)
        ]
    )


class TestTimeSeriesIO:
    @pytest.mark.usefixtures("disable_gzip", "disable_pypi_check")
    def test_download_upload(
        self,
        toolkit_config: ToolkitClientConfig,
        some_timeseries_data: TimeSeriesList,
        respx_mock: respx.MockRouter,
    ) -> None:
        config = toolkit_config
        ts_by_external_id = {ts.external_id: ts for ts in some_timeseries_data if ts.external_id is not None}
        selector = DataSetSelector(data_set_external_id="DataSetSelector", kind="TimeSeries")

        def create_callback(request: httpx.Request) -> httpx.Response:
            payload = json.loads(request.content)
            assert "items" in payload
            items = payload["items"]
            assert isinstance(items, list)
            return httpx.Response(
                status_code=200,
                json={
                    "items": [ts_by_external_id[item["externalId"]].dump() for item in items if "externalId" in item]
                },
            )

        respx_mock.post(config.create_api_url("/timeseries")).mock(side_effect=create_callback)
        with monkeypatch_toolkit_client() as client:
            client.config = config
            client.time_series.return_value = chunker(some_timeseries_data, 10)
            client.time_series.aggregate_count.return_value = len(some_timeseries_data)
            client.lookup.data_sets.external_id.return_value = "test_data_set"
            client.lookup.data_sets.id.return_value = 1234
            client.lookup.assets.external_id.return_value = "test_hierarchy"
            client.lookup.assets.id.return_value = 123

            io = TimeSeriesIO(client)

            assert io.count(selector) == 50

            source = io.stream_data(selector)
            json_chunks: list[list[dict[str, JsonVal]]] = []
            for page in source:
                # New interface: stream_data returns Page objects
                json_chunk = io.data_to_json_chunk(page.items)
                assert isinstance(json_chunk, list)
                assert len(json_chunk) == 10
                for item in json_chunk:
                    assert isinstance(item, dict)
                    assert "dataSetExternalId" in item
                    assert item["dataSetExternalId"] == "test_data_set"
                json_chunks.append(json_chunk)

            with HTTPClient(config) as upload_client:
                from cognite_toolkit._cdf_tk.storageio._base import UploadItem

                # New interface: convert individual items and create UploadItems
                for chunk in json_chunks:
                    write_items = [io.json_to_resource(item) for item in chunk]
                    upload_items = [UploadItem(source_id=io.as_id(item), item=item) for item in write_items]
                    io.upload_items(upload_items, upload_client, selector)

            assert respx_mock.calls.call_count == 5  # 50 rows in chunks of 10
            uploaded_ts = []
            for call in respx_mock.calls:
                uploaded_ts.extend(json.loads(call.request.content)["items"])

            assert uploaded_ts == some_timeseries_data.as_write().dump()


@pytest.fixture()
def some_event_data() -> EventList:
    """Fixture to provide a sample EventList for testing."""
    return EventList(
        [
            Event(
                external_id=f"event_{i}",
                description=f"Description for event {i}",
                asset_ids=[123],
                data_set_id=1234,
                source="test_source",
                start_time=1000000000000 + i * 1000,
                end_time=1000000001000 + i * 1000,
            )
            for i in range(50)
        ]
    )


class TestEventIO:
    @pytest.mark.usefixtures("disable_gzip", "disable_pypi_check")
    def test_download_upload(
        self,
        toolkit_config: ToolkitClientConfig,
        some_event_data: EventList,
        respx_mock: respx.MockRouter,
    ) -> None:
        config = toolkit_config
        event_by_external_id = {event.external_id: event for event in some_event_data if event.external_id is not None}
        selector = DataSetSelector(data_set_external_id="DataSetSelector", kind="Events")

        def create_callback(request: httpx.Request) -> httpx.Response:
            payload = json.loads(request.content)
            assert "items" in payload
            items = payload["items"]
            assert isinstance(items, list)
            return httpx.Response(
                status_code=200,
                json={
                    "items": [event_by_external_id[item["externalId"]].dump() for item in items if "externalId" in item]
                },
            )

        respx_mock.post(config.create_api_url("/events")).mock(side_effect=create_callback)
        with monkeypatch_toolkit_client() as client:
            client.config = config
            client.events.return_value = chunker(some_event_data, 10)
            client.events.aggregate_count.return_value = len(some_event_data)
            client.lookup.data_sets.external_id.return_value = "test_data_set"
            client.lookup.data_sets.id.return_value = 1234
            client.lookup.assets.external_id.return_value = ["test_hierarchy"]
            client.lookup.assets.id.return_value = [123]

            io = EventIO(client)

            assert io.count(selector) == 50

            source = io.stream_data(selector)
            json_chunks: list[list[dict[str, JsonVal]]] = []
            for page in source:
                json_chunk = io.data_to_json_chunk(page.items)
                assert isinstance(json_chunk, list)
                assert len(json_chunk) == 10
                for item in json_chunk:
                    assert isinstance(item, dict)
                    assert "dataSetExternalId" in item
                    assert item["dataSetExternalId"] == "test_data_set"
                json_chunks.append(json_chunk)

            with HTTPClient(config) as upload_client:
                from cognite_toolkit._cdf_tk.storageio._base import UploadItem

                # New interface: convert individual items and create UploadItems
                for chunk in json_chunks:
                    write_items = [io.json_to_resource(item) for item in chunk]
                    upload_items = [UploadItem(source_id=io.as_id(item), item=item) for item in write_items]
                    io.upload_items(upload_items, upload_client, selector)

            assert respx_mock.calls.call_count == 5  # 50 rows in chunks of 10
            uploaded_events = []
            for call in respx_mock.calls:
                uploaded_events.extend(json.loads(call.request.content)["items"])

            assert uploaded_events == some_event_data.as_write().dump()



import json
from collections.abc import Hashable, Iterable, Sequence
from itertools import zip_longest
from pathlib import Path
from typing import Any, Literal, cast, final

from cognite.client.data_classes import (
    DatapointSubscription,
    DatapointSubscriptionList,
    DataPointSubscriptionUpdate,
    DataPointSubscriptionWrite,
    DatapointSubscriptionWriteList,
    TimeSeries,
    TimeSeriesList,
    TimeSeriesWrite,
    TimeSeriesWriteList,
)
from cognite.client.data_classes.capabilities import (
    Capability,
    TimeSeriesAcl,
    TimeSeriesSubscriptionsAcl,
)
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.data_classes.datapoints_subscriptions import TimeSeriesIDList
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.constants import MAX_TIMESTAMP_MS, MIN_TIMESTAMP_MS
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceContainerCRUD, ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitRequiredValueError,
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.resource_classes import DatapointSubscriptionYAML, TimeSeriesYAML
from cognite_toolkit._cdf_tk.utils import calculate_hash
from cognite_toolkit._cdf_tk.utils.collection import chunker
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable, dm_identifier
from cognite_toolkit._cdf_tk.utils.text import suffix_description

from .auth import GroupAllScopedCRUD, SecurityCategoryCRUD
from .classic import AssetCRUD
from .data_organization import DataSetsCRUD


@final
class TimeSeriesCRUD(ResourceContainerCRUD[str, TimeSeriesWrite, TimeSeries, TimeSeriesWriteList, TimeSeriesList]):
    item_name = "datapoints"
    folder_name = "timeseries"
    filename_pattern = r"^(?!.*DatapointSubscription$).*"
    resource_cls = TimeSeries
    resource_write_cls = TimeSeriesWrite
    list_cls = TimeSeriesList
    list_write_cls = TimeSeriesWriteList
    yaml_cls = TimeSeriesYAML
    kind = "TimeSeries"
    dependencies = frozenset({DataSetsCRUD, GroupAllScopedCRUD, AssetCRUD})
    _doc_url = "Time-series/operation/postTimeSeries"

    @property
    def display_name(self) -> str:
        return "time series"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[TimeSeriesWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = [TimeSeriesAcl.Action.Read] if read_only else [TimeSeriesAcl.Action.Read, TimeSeriesAcl.Action.Write]

        scope: TimeSeriesAcl.Scope.All | TimeSeriesAcl.Scope.DataSet = TimeSeriesAcl.Scope.All()  # type: ignore[valid-type]
        if items:
            if dataset_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = TimeSeriesAcl.Scope.DataSet(list(dataset_ids))

        return TimeSeriesAcl(actions, scope)

    @classmethod
    def get_id(cls, item: TimeSeries | TimeSeriesWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if item.external_id is None:
            raise ToolkitRequiredValueError("TimeSeries must have external_id set.")
        return item.external_id

    @classmethod
    def get_internal_id(cls, item: TimeSeries | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        return item.id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, item["dataSetExternalId"]
        if "securityCategoryNames" in item:
            for security_category in item["securityCategoryNames"]:
                yield SecurityCategoryCRUD, security_category
        if "assetExternalId" in item:
            yield AssetCRUD, item["assetExternalId"]

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> TimeSeriesWrite:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if security_categories_names := resource.pop("securityCategoryNames", []):
            resource["securityCategories"] = self.client.lookup.security_categories.id(
                security_categories_names, is_dry_run
            )
        if asset_external_id := resource.pop("assetExternalId", None):
            resource["assetId"] = self.client.lookup.assets.id(asset_external_id, is_dry_run)
        return TimeSeriesWrite._load(resource)

    def dump_resource(self, resource: TimeSeries, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if security_categories := dumped.pop("securityCategories", []):
            dumped["securityCategoryNames"] = self.client.lookup.security_categories.external_id(security_categories)
        if asset_id := dumped.pop("assetId", None):
            dumped["assetExternalId"] = self.client.lookup.assets.external_id(asset_id)
        return dumped

    def create(self, items: TimeSeriesWriteList) -> TimeSeriesList:
        return self.client.time_series.create(items)

    def retrieve(self, ids: SequenceNotStr[str | int]) -> TimeSeriesList:
        internal_ids, external_ids = self._split_ids(ids)
        return self.client.time_series.retrieve_multiple(
            ids=internal_ids, external_ids=external_ids, ignore_unknown_ids=True
        )

    def update(self, items: TimeSeriesWriteList) -> TimeSeriesList:
        return self.client.time_series.update(items, mode="replace")

    def delete(self, ids: SequenceNotStr[str | int]) -> int:
        existing = self.retrieve(ids)
        if existing:
            self.client.time_series.delete(id=existing.as_ids(), ignore_unknown_ids=True)
        return len(existing)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[TimeSeries]:
        return iter(
            self.client.time_series(data_set_external_ids=[data_set_external_id] if data_set_external_id else None)
        )

    def count(self, ids: str | dict[str, Any] | SequenceNotStr[str | dict[str, Any]] | None) -> int:
        datapoints = self.client.time_series.data.retrieve(
            external_id=ids,  # type: ignore[arg-type]
            start=MIN_TIMESTAMP_MS,
            end=MAX_TIMESTAMP_MS + 1,
            aggregates="count",
            granularity="1000d",
            ignore_unknown_ids=True,
        )
        return sum(sum(data.count or []) for data in datapoints)  # type: ignore[union-attr, misc, arg-type]

    def drop_data(self, ids: SequenceNotStr[str] | None) -> int:
        count = self.count(ids)
        existing = self.client.time_series.retrieve_multiple(
            external_ids=cast(SequenceNotStr[str], ids), ignore_unknown_ids=True
        ).as_external_ids()
        for external_id in existing:
            self.client.time_series.data.delete_range(
                external_id=external_id, start=MIN_TIMESTAMP_MS, end=MAX_TIMESTAMP_MS + 1
            )
        return count


@final
class DatapointSubscriptionCRUD(
    ResourceCRUD[
        str,
        DataPointSubscriptionWrite,
        DatapointSubscription,
        DatapointSubscriptionWriteList,
        DatapointSubscriptionList,
    ]
):
    folder_name = "timeseries"
    filename_pattern = r"^.*DatapointSubscription$"  # Matches all yaml files who end with *DatapointSubscription.
    resource_cls = DatapointSubscription
    resource_write_cls = DataPointSubscriptionWrite
    list_cls = DatapointSubscriptionList
    list_write_cls = DatapointSubscriptionWriteList
    kind = "DatapointSubscription"
    _doc_url = "Data-point-subscriptions/operation/postSubscriptions"
    dependencies = frozenset(
        {
            TimeSeriesCRUD,
            GroupAllScopedCRUD,
        }
    )
    yaml_cls = DatapointSubscriptionYAML

    _hash_key = "cdf-hash"
    _description_character_limit = 1000
    # A datapoint subscription can hold 10,000 timeseries, but the API
    # only supports 100 timeseries per request. Thus, if a subscription
    # has more than 100 timeseries, we need to split it into multiple requests.
    _TIMESERIES_ID_REQUEST_LIMIT = 100
    _MAX_TIMESERIES_IDS = 10_000

    @property
    def display_name(self) -> str:
        return "timeseries subscriptions"

    @classmethod
    def get_id(cls, item: DataPointSubscriptionWrite | DatapointSubscription | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, item["dataSetExternalId"]
        for timeseries_id in item.get("timeSeriesIds", []):
            yield TimeSeriesCRUD, timeseries_id

    @classmethod
    def get_required_capability(
        cls, items: Sequence[DataPointSubscriptionWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [TimeSeriesSubscriptionsAcl.Action.Read]
            if read_only
            else [TimeSeriesSubscriptionsAcl.Action.Read, TimeSeriesSubscriptionsAcl.Action.Write]
        )

        scope: TimeSeriesSubscriptionsAcl.Scope.All | TimeSeriesSubscriptionsAcl.Scope.DataSet = (  # type: ignore[valid-type]
            TimeSeriesSubscriptionsAcl.Scope.All()
        )
        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = TimeSeriesSubscriptionsAcl.Scope.DataSet(list(data_set_ids))

        return TimeSeriesSubscriptionsAcl(actions, scope)

    def create(self, items: DatapointSubscriptionWriteList) -> DatapointSubscriptionList:
        created_list = DatapointSubscriptionList([])
        for item in items:
            to_create, batches = self.create_split_timeseries_ids(item)
            created = self.client.time_series.subscriptions.create(to_create)
            for batch_item in batches:
                created = self.client.time_series.subscriptions.update(batch_item)
            created_list.append(created)
        return created_list

    def retrieve(self, ids: SequenceNotStr[str]) -> DatapointSubscriptionList:
        items = DatapointSubscriptionList([])
        for id_ in ids:
            retrieved = self.client.time_series.subscriptions.retrieve(id_)
            if retrieved:
                items.append(retrieved)
        return items

    def update(self, items: DatapointSubscriptionWriteList) -> DatapointSubscriptionList:
        updated_list = DatapointSubscriptionList([])
        for item in items:
            current = self.client.time_series.subscriptions.list_member_time_series(item.external_id, limit=-1)
            to_update, batches = self.update_split_timeseries_ids(item, current)
            # There are two versions of a TimeSeries Subscription, one selects timeseries based filter
            # and the other selects timeseries based on timeSeriesIds. If we use mode='replace', we try
            # to set timeSeriesIds to an empty list, while the filter is set. This will result in an error.
            updated = self.client.time_series.subscriptions.update(to_update, mode="replace_ignore_null")
            for batch_item in batches:
                updated = self.client.time_series.subscriptions.update(batch_item)
            updated_list.append(updated)

        return updated_list

    def delete(self, ids: SequenceNotStr[str]) -> int:
        try:
            self.client.time_series.subscriptions.delete(ids)
        except (CogniteAPIError, CogniteNotFoundError) as e:
            non_existing = set(e.failed or [])
            if existing := [id_ for id_ in ids if id_ not in non_existing]:
                self.client.time_series.subscriptions.delete(existing)
            return len(existing)
        else:
            # All deleted successfully
            return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[DatapointSubscription]:
        return iter(self.client.time_series.subscriptions)

    def load_resource_file(
        self, filepath: Path, environment_variables: dict[str, str | None] | None = None
    ) -> list[dict[str, Any]]:
        resources = super().load_resource_file(filepath, environment_variables)
        for resource in resources:
            if "timeSeriesIds" not in resource and "instanceIds" not in resource:
                continue
            # If the timeSeriesIds or instanceIds is set, we need to add the auth hash to the description.
            # such that we can detect if the subscription has changed.
            content: dict[str, object] = {}
            if "timeSeriesIds" in resource:
                content["timeSeriesIds"] = resource["timeSeriesIds"]
            if "instanceIds" in resource:
                content["instanceIds"] = resource["instanceIds"]
            timeseries_hash = calculate_hash(json.dumps(content), shorten=True)
            extra_str = f"{self._hash_key}: {timeseries_hash}"
            resource["description"] = suffix_description(
                extra_str,
                resource.get("description"),
                self._description_character_limit,
                self.get_id(resource),
                self.display_name,
                self.console,
            )

        return resources

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> DataPointSubscriptionWrite:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        return DataPointSubscriptionWrite._load(resource)

    def dump_resource(self, resource: DatapointSubscription, local: dict[str, Any] | None = None) -> dict[str, Any]:
        if resource.filter is not None:
            dumped = resource.as_write().dump()
        else:
            # If filter is not set, the subscription uses explicit timeSeriesIds, which are not returned in the
            # response. Calling .as_write() in this case raises ValueError because either filter or
            # timeSeriesIds must be set.
            dumped = resource.dump()
            for server_prop in ("createdTime", "lastUpdatedTime", "timeSeriesCount"):
                dumped.pop(server_prop, None)
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        # timeSeriesIds and instanceIds are not returned in the response, so we need to add them
        # to the dumped resource if they are set in the local resource. If there is a discrepancy between
        # the local and dumped resource, th hash added to the description will change.
        if "timeSeriesIds" in local:
            dumped["timeSeriesIds"] = local["timeSeriesIds"]
        if "instanceIds" in local:
            dumped["instanceIds"] = local["instanceIds"]
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path[0] == "filter" or json_path == ("timeSeriesIds",):
            return diff_list_hashable(local, cdf)
        elif json_path == ("instanceIds",):
            return diff_list_identifiable(local, cdf, get_identifier=dm_identifier)
        return super().diff_list(local, cdf, json_path)

    @classmethod
    def create_split_timeseries_ids(
        cls, subscription: DataPointSubscriptionWrite
    ) -> tuple[DataPointSubscriptionWrite, list[DataPointSubscriptionUpdate]]:
        """Split the time series IDs into batches of 100.
        This is needed because the API only supports 100 time series IDs per request.

        Note this is the total of time series IDs and instance IDs.
        """
        total_timeseries = len(subscription.time_series_ids or []) + len(subscription.instance_ids or [])
        cls._validate_total_below_limit(subscription, total_timeseries)
        if total_timeseries <= cls._TIMESERIES_ID_REQUEST_LIMIT:
            # If the subscription has less than or equal to 100 time series IDs, we can return it as is.
            # No need to split into batches.
            return subscription, []

        # Serialization to create a copy of the subscription
        to_create = DataPointSubscriptionWrite.load(subscription.dump())
        all_timeseries_ids = to_create.time_series_ids or []
        all_instance_ids = to_create.instance_ids or []

        # Create the first batch for the create/update call, prioritizing time_series_ids
        to_create.time_series_ids = all_timeseries_ids[: cls._TIMESERIES_ID_REQUEST_LIMIT]
        space_in_first_batch = cls._TIMESERIES_ID_REQUEST_LIMIT - len(to_create.time_series_ids)
        to_create.instance_ids = all_instance_ids[:space_in_first_batch]

        # Prepare remaining IDs for update batches
        remaining_timeseries_ids = all_timeseries_ids[len(to_create.time_series_ids) :]
        remaining_instance_ids = all_instance_ids[len(to_create.instance_ids) :]

        # Using a list of tuples to preserve the type of ID
        all_remaining_ids = [("ts", id) for id in remaining_timeseries_ids] + [
            ("instance", id) for id in remaining_instance_ids
        ]

        batches: list[DataPointSubscriptionUpdate] = []
        for chunk in chunker(all_remaining_ids, cls._TIMESERIES_ID_REQUEST_LIMIT):
            update = DataPointSubscriptionUpdate(external_id=subscription.external_id)
            ts_ids_in_chunk, instance_ids_in_chunk = cls._split_ts_instance_ids(chunk)
            if ts_ids_in_chunk:
                update.time_series_ids.add(ts_ids_in_chunk)
            if instance_ids_in_chunk:
                update.instance_ids.add(instance_ids_in_chunk)
            batches.append(update)
        return to_create, batches

    @classmethod
    def _validate_total_below_limit(cls, subscription: DataPointSubscriptionWrite, total_timeseries: int) -> None:
        if total_timeseries > cls._MAX_TIMESERIES_IDS:
            raise ToolkitValueError(
                f'Subscription "{subscription.external_id}" has {total_timeseries:,} time series, '
                f"which is more than the limit of {cls._MAX_TIMESERIES_IDS:,}."
            )

    @classmethod
    def _split_ts_instance_ids(
        cls, ids: list[tuple[Literal["ts"], str] | tuple[Literal["instance"], NodeId]]
    ) -> tuple[list[str], list[NodeId]]:
        ts_ids, instance_ids = [], []
        for id_type, identifier in ids:
            if id_type == "ts":
                ts_ids.append(identifier)
            else:
                instance_ids.append(identifier)
        # MyPy fails to understand the logic above ensures that ts_ids is a list of str
        # and instance_ids is a list of NodeId.
        return ts_ids, instance_ids  # type: ignore[return-value]

    @classmethod
    def update_split_timeseries_ids(
        cls, subscription: DataPointSubscriptionWrite, current_ts: TimeSeriesIDList
    ) -> tuple[DataPointSubscriptionWrite, list[DataPointSubscriptionUpdate]]:
        """Split the time series IDs into batches of 100.
        This is needed because the API only supports 100 time series IDs per request.

        Note this is the total of time series IDs and instance IDs.

        In addition, this method will compare to the current times series IDs and
        update the subscription with adding and removing time series IDs as needed.
        """
        if subscription.time_series_ids is None and subscription.instance_ids is None:
            # The subscription is using a filter, so we can return it as is.
            return subscription, []
        total_timeseries = len(subscription.time_series_ids or []) + len(subscription.instance_ids or [])
        cls._validate_total_below_limit(subscription, total_timeseries)

        # Serialization to create a copy of the subscription
        to_update = DataPointSubscriptionWrite.load(subscription.dump())

        # Get desired time series IDs
        desired_timeseries_ids = set(to_update.time_series_ids or [])
        desired_instance_ids = set(to_update.instance_ids or [])

        # Get current time series IDs from the subscription
        current_timeseries_ids: set[str] = set()
        current_instance_ids: set[NodeId] = set()
        for ts in current_ts:
            if ts.external_id and ts.instance_id is None:
                current_timeseries_ids.add(ts.external_id)
            elif ts.instance_id and ts.external_id is None:
                current_instance_ids.add(ts.instance_id)
            else:
                raise ValueError(f"External ID and instance ID are both set for time series {ts.external_id}.")

        # Calculate what needs to be added and removed
        ts_to_add = desired_timeseries_ids - current_timeseries_ids
        ts_to_remove = current_timeseries_ids - desired_timeseries_ids
        instance_to_add = desired_instance_ids - current_instance_ids
        instance_to_remove = current_instance_ids - desired_instance_ids

        # Clear the time series IDs from the main update to avoid conflicts
        # The to_update object is used to update all other properties of the subscription.
        to_update.time_series_ids = None
        to_update.instance_ids = None

        # Create update batches for changes
        batches: list[DataPointSubscriptionUpdate] = []
        all_removals = [("ts", id_) for id_ in ts_to_remove] + [("instance", id_) for id_ in instance_to_remove]
        all_additions = [("ts", id_) for id_ in ts_to_add] + [("instance", id_) for id_ in instance_to_add]
        for removals, additions in zip_longest(
            chunker(all_removals, cls._TIMESERIES_ID_REQUEST_LIMIT),
            chunker(all_additions, cls._TIMESERIES_ID_REQUEST_LIMIT),
            fillvalue=None,
        ):
            update = DataPointSubscriptionUpdate(external_id=subscription.external_id)
            ts_ids_to_remove, instance_ids_to_remove = cls._split_ts_instance_ids(removals or [])
            if ts_ids_to_remove:
                update.time_series_ids.remove(ts_ids_to_remove)
            if instance_ids_to_remove:
                update.instance_ids.remove(instance_ids_to_remove)

            ts_ids_to_add, instance_ids_to_add = cls._split_ts_instance_ids(additions or [])
            if ts_ids_to_add:
                update.time_series_ids.add(ts_ids_to_add)
            if instance_ids_to_add:
                update.instance_ids.add(instance_ids_to_add)
            batches.append(update)

        return to_update, batches
